import pytest
from saltfactories.utils import random_string

from tests.support.mysql import MySQLImage
from tests.support.mysql import create_mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_container  # pylint: disable=unused-import
from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_revoke
from tests.support.vault import vault_write

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "cache": {
                "backend": "disk",  # ensure a persistent cache is available for get_creds
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        }
    }


@pytest.fixture(scope="module")
def mysql_image():
    version = "10.3"
    return MySQLImage(
        name="mariadb",
        tag=version,
        container_id=random_string(f"mariadb-{version}-"),
    )


@pytest.fixture
def role_args_common():
    return {
        "db_name": "testdb",
        "creation_statements": r"CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}';GRANT SELECT ON *.* TO '{{name}}'@'%';",
    }


@pytest.fixture
def testrole():
    return {
        "default_ttl": 3600,
        "max_ttl": 7200,
    }


@pytest.fixture
def testdb(mysql_container, container_host_ref):
    return {
        "plugin_name": "mysql-database-plugin",
        "connection_url": f"{{{{username}}}}:{{{{password}}}}@tcp({container_host_ref}:{mysql_container.mysql_port})/",
        "allowed_roles": "testrole,teststaticrole,testreissuerole",
        "username": "root",
        "password": mysql_container.mysql_passwd,
    }


@pytest.fixture(scope="module", autouse=True)
def db_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("database")
    yield
    assert vault_disable_secret_engine("database")


@pytest.fixture
def connection_setup(testdb):
    try:
        vault_write("database/config/testdb", **testdb)
        assert "testdb" in vault_list("database/config")
        yield
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")


@pytest.fixture(params=[["testrole"]])
def roles_setup(connection_setup, request, role_args_common):  # pylint: disable=unused-argument
    try:
        for role_name in request.param:
            role_args = request.getfixturevalue(role_name)
            role_args.update(role_args_common)
            vault_write(f"database/roles/{role_name}", **role_args)
            assert role_name in vault_list("database/roles")
        yield
    finally:
        for role_name in request.param:
            if role_name in vault_list("database/roles"):
                vault_delete(f"database/roles/{role_name}")
                assert role_name not in vault_list("database/roles")


@pytest.fixture
def vault_db(modules):
    try:
        yield modules.vault_db
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")
        if "testrole" in vault_list("database/roles"):
            vault_delete("database/roles/testrole")
            assert "testrole" not in vault_list("database/roles")


@pytest.fixture(params=({},))
def lease_creation_params(request):
    defaults = {"name": "testrole"}
    defaults.update(request.param)
    return defaults


@pytest.fixture(params=(False,))
def _multi_lease(request):
    return request.param


@pytest.fixture(params=({},))
def beacon_config(request, _multi_lease, existing_lease):
    req = request.param.copy()
    datatype = req.pop("leases_type", str if not _multi_lease else list)
    if _multi_lease:
        alt_lease = request.getfixturevalue("existing_alt_lease")
    data = {}
    if datatype is str:
        data["leases"] = existing_lease
    elif datatype is list:
        data["leases"] = [existing_lease]
        if _multi_lease:
            data["leases"].append(alt_lease)
    else:
        data["leases"] = {existing_lease: req.pop("per_lease_params", {})}
        if _multi_lease:
            data["leases"][alt_lease] = req.pop("per_lease_alt_params", {})
    data.update(req)
    return [{k: v} for k, v in data.items()]


@pytest.fixture
def existing_lease(
    roles_setup, lease_creation_params, vault_db, loaders
):  # pylint: disable=unused-argument
    ckey = ".".join(
        [
            "db",
            lease_creation_params.get("mount", "database"),
            "dynamic",
            lease_creation_params["name"],
            lease_creation_params.get("cache", "default"),
        ]
    )
    lease = vault_db.get_creds(**lease_creation_params)
    assert lease
    # We need to clear the context because in the test suite, the beacon modules
    # are running in a different one than the execution modules and the lease
    # has already been cached in the context of the execution module.
    # This means it does not pick up changes to the cached files, but we need
    # it to check changes in the tests.
    loaders.context.clear()
    return ckey  # revocation is handled in vault_db


@pytest.fixture(params=({"cache": "alt"},))
def existing_alt_lease(
    request, roles_setup, lease_creation_params, vault_db, loaders
):  # pylint: disable=unused-argument
    params = request.param
    ckey = ".".join(
        [
            "db",
            params.get("mount", lease_creation_params.get("mount", "database")),
            "dynamic",
            params.get("name", lease_creation_params["name"]),
            params.get("cache", lease_creation_params.get("cache", "default")),
        ]
    )
    lease = vault_db.get_creds(**lease_creation_params, **params)
    assert lease
    # We need to clear the context because in the test suite, the beacon modules
    # are running in a different one than the execution modules and the lease
    # has already been cached in the context of the execution module.
    # This means it does not pick up changes to the cached files, but we need
    # it to check changes in the tests.
    loaders.context.clear()
    return ckey  # revocation is handled in vault_db


@pytest.fixture
def revoked_lease(existing_lease, vault_db):
    lease_id = vault_db.list_cached()[existing_lease]["lease_id"]
    assert vault_revoke(lease_id)
    return existing_lease


@pytest.fixture
def beacon(beacons):
    yield beacons.vault_lease.beacon


@pytest.mark.usefixtures("existing_alt_lease")
@pytest.mark.usefixtures("_multi_lease")
@pytest.mark.parametrize("_multi_lease", (False, True), indirect=True)
@pytest.mark.parametrize(
    "beacon_config", ({"check_server": False}, {"check_server": True}), indirect=True
)
def test_beacon_valid(beacon, beacon_config):
    ret = beacon(beacon_config)
    assert ret == []


@pytest.mark.parametrize("beacon_config", ({"leases": "foo.bar.baz"},), indirect=True)
def test_beacon_missing(beacon, beacon_config):
    ret = beacon(beacon_config)
    assert ret == [
        {
            "check_server": None,
            "ckey": "foo.bar.baz",
            "expired": True,
            "expires_in": -1,
            "meta": None,
            "min_ttl": 300,
            "tag": "expire",
        }
    ]


@pytest.mark.parametrize("beacon_config", ({"leases": ["foo.bar", "foo.baz"]},), indirect=True)
def test_beacon_missing_multi(beacon, beacon_config):
    ret = beacon(beacon_config)
    assert ret == [
        {
            "check_server": None,
            "ckey": "foo.bar",
            "expired": True,
            "expires_in": -1,
            "meta": None,
            "min_ttl": 300,
            "tag": "expire",
        },
        {
            "check_server": None,
            "ckey": "foo.baz",
            "expired": True,
            "expires_in": -1,
            "meta": None,
            "min_ttl": 300,
            "tag": "expire",
        },
    ]


@pytest.mark.usefixtures("existing_alt_lease")
@pytest.mark.usefixtures("_multi_lease")
@pytest.mark.parametrize("_multi_lease", (True,), indirect=True)
@pytest.mark.usefixtures("revoked_lease")
def test_beacon_revoked_not_check_server(beacon, beacon_config):
    ret = beacon(beacon_config)
    assert ret == []


@pytest.mark.usefixtures("existing_alt_lease")
@pytest.mark.usefixtures("_multi_lease")
@pytest.mark.parametrize("_multi_lease", (True,), indirect=True)
@pytest.mark.usefixtures("revoked_lease", "lease_creation_params")
@pytest.mark.parametrize(
    "beacon_config,lease_creation_params",
    (
        ({"check_server": True}, {}),
        ({"leases_type": dict, "per_lease_params": {"check_server": True}}, {}),
    ),
    indirect=True,
)
def test_beacon_revoked_check_server(beacon, beacon_config):
    ret = beacon(beacon_config)
    assert len(ret) == 1
    ret = ret[0]
    _assert_evt(ret, check_server=True, expired=True, expires_in=-1)


@pytest.mark.usefixtures("beacon_config", "lease_creation_params")
@pytest.mark.parametrize(
    "beacon_config,lease_creation_params",
    (
        ({"min_ttl": 7000}, {}),
        ({"leases_type": dict, "per_lease_params": {"min_ttl": 7000}}, {}),
        ({}, {"valid_for": 7000}),
        ({"min_ttl": 300}, {"valid_for": 7000}),
        ({"leases_type": dict, "per_lease_params": {"min_ttl": 300}}, {"valid_for": 7000}),
    ),
    indirect=True,
)
def test_beacon_min_ttl(beacon, beacon_config, vault_db, existing_lease):
    ret = beacon(beacon_config)
    assert ret == []
    info = vault_db.list_cached()[existing_lease]
    assert info["duration"] == 7000


@pytest.mark.usefixtures("beacon_config", "existing_lease")
@pytest.mark.parametrize(
    "beacon_config",
    ({"renew": False, "min_ttl": 7000},),
    indirect=True,
)
def test_beacon_not_renew(beacon, beacon_config):
    ret = beacon(beacon_config)
    assert len(ret) == 1
    ret = ret[0]
    _assert_evt(ret, min_ttl=7000)


@pytest.mark.usefixtures("beacon_config", "lease_creation_params")
@pytest.mark.parametrize(
    "beacon_config,lease_creation_params",
    (
        ({"min_ttl": 8000}, {}),
        ({"leases_type": dict, "per_lease_params": {"min_ttl": 8000}}, {}),
        ({}, {"valid_for": 8000}),
        ({"min_ttl": 300}, {"valid_for": 8000}),
        ({"leases_type": dict, "per_lease_params": {"min_ttl": 300}}, {"valid_for": 8000}),
    ),
    indirect=True,
)
def test_beacon_min_ttl_unattainable(beacon, beacon_config):
    ret = beacon(beacon_config)
    assert len(ret) == 1
    ret = ret[0]
    _assert_evt(ret, min_ttl=8000)
    assert ret["expires_in"] > 3590


@pytest.mark.usefixtures("beacon_config", "lease_creation_params")
@pytest.mark.parametrize(
    "beacon_config,lease_creation_params,expected_meta",
    (
        ({"meta": "foo.bar"}, {}, "foo.bar"),
        ({"leases_type": dict, "per_lease_params": {"meta": "foo.bar"}}, {}, "foo.bar"),
        ({}, {"meta": "foo.bar"}, "foo.bar"),
        ({"meta": "foo.bar"}, {"meta": "foo.baz"}, "foo.baz"),
        (
            {"leases_type": dict, "per_lease_params": {"meta": "foo.bar"}},
            {"meta": "foo.baz"},
            "foo.baz",
        ),
        ({"meta": "foo.bar"}, {"meta": ["foo.baz"]}, ["foo.baz"]),
        ({"meta": ["foo.bar"]}, {"meta": "foo.baz"}, "foo.baz"),
        ({"meta": ["foo.bar"]}, {"meta": {"foo": "baz"}}, {"foo": "baz"}),
        ({"meta": ["foo.bar"]}, {"meta": ["foo.baz"]}, ["foo.bar", "foo.baz"]),
        (
            {"meta": {"foo": {"bar": True}}},
            {"meta": {"foo": {"bar": False}}},
            {"foo": {"bar": False}},
        ),
        (
            {"meta": {"foo": {"bar": True}}},
            {"meta": {"foo": {"baz": False}}},
            {"foo": {"bar": True, "baz": False}},
        ),
        (
            {"meta": {"foo": {"bar": [True]}}},
            {"meta": {"foo": {"bar": [False]}}},
            {"foo": {"bar": [True, False]}},
        ),
        ({"meta": "foo"}, {"meta": {"foo": {"bar": False}}}, {"foo": {"bar": False}}),
    ),
    indirect=("beacon_config", "lease_creation_params"),
)
def test_beacon_meta(beacon, beacon_config, expected_meta):
    beacon_config.append({"min_ttl": 10000})
    ret = beacon(beacon_config)
    assert len(ret) == 1
    ret = ret[0]
    _assert_evt(ret, min_ttl=10000, meta=expected_meta)


def _assert_evt(evt, *remove, **expected):
    assert set(evt) == {
        "meta",
        "creation_time",
        "duration",
        "expired",
        "revoke_delay",
        "tag",
        "renew_increment",
        "renewable",
        "min_ttl",
        "check_server",
        "lease_id",
        "ckey",
        "expire_time",
        "expires_in",
    }
    expected = {
        "check_server": None,
        "ckey": "db.database.dynamic.testrole.default",
        # For renewals, we can't know what the max_ttl is, so this will be the default
        # duration.
        "duration": 3600,
        "expired": False,
        "meta": None,
        "min_ttl": 300,
        "renew_increment": None,
        "renewable": True,
        "revoke_delay": None,
        "tag": "expire",
        **expected,
    }
    expected.update(expected)
    for unwanted in remove:
        expected.pop(unwanted, None)
    for param, val in expected.items():
        assert evt[param] == val



"""
:copyright: Copyright 2013-2017 by the SaltStack Team, see AUTHORS for more details.
:license: Apache 2.0, see LICENSE for more details.


tests.support.helpers
~~~~~~~~~~~~~~~~~~~~~

Test support helpers
"""

import logging
import os

log = logging.getLogger(__name__)


class PatchedEnviron:
    """
    Create a patched environment
    """

    def __init__(self, **kwargs):
        self.cleanup_keys = kwargs.pop("__cleanup__", ())
        self.kwargs = kwargs
        self.original_environ = None

    def __enter__(self):
        self.original_environ = os.environ.copy()
        for key in self.cleanup_keys:
            os.environ.pop(key, None)
        os.environ.update(**self.kwargs)
        return self

    def __exit__(self, *args):
        os.environ.clear()
        os.environ.update(self.original_environ)



"""
Manage Vault (or OpenBao) KV v1/v2 secrets statefully.

.. versionadded:: 1.2.0

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import copy
import logging

from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltException
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def present(name, values, sync=False):
    """
    Ensure a secret is present as specified.
    Does not report a diff.

    name
        The path of the secret.

    values
        A mapping of values the secret should expose.

    sync
        Ensure the secret only exposes ``values`` and delete unspecified ones.
        Defaults to false, which results in patching (merging over) existing data
        and deleting keys that are set to ``None``/``null``. For details, see
        https://datatracker.ietf.org/doc/html/draft-ietf-appsawg-json-merge-patch-07
    """
    # TODO: manage KV v2 metadata?
    ret = {
        "name": name,
        "result": True,
        "comment": "The secret is already present as specified",
        "changes": {},
    }
    try:
        try:
            current = __salt__["vault.read_secret"](name)
        except CommandExecutionError as err:
            # VaultNotFoundError should be subclassed to
            # CommandExecutionError and not re-raised by the
            # execution module @FIXME?
            if "VaultNotFoundError" not in str(err):
                raise
            current = None
        else:
            if sync:
                if current == values:
                    return ret
            else:

                def apply_json_merge_patch(data, patch):
                    if not patch:
                        return data
                    if not isinstance(data, dict) or not isinstance(patch, dict):
                        raise ValueError("Data and patch must be dictionaries.")

                    for key, value in patch.items():
                        if value is None:
                            data.pop(key, None)
                        elif isinstance(value, dict):
                            data[key] = apply_json_merge_patch(data.get(key, {}), value)
                        else:
                            data[key] = value
                    return data

                new = apply_json_merge_patch(copy.deepcopy(current), values)
                if new == current:
                    return ret
        verb = "patch" if current is not None and not sync else "write"
        pp = "patched" if verb == "patch" else "written"
        ret["changes"][pp] = name
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Would have {pp} the secret"
            return ret
        if not __salt__[f"vault.{verb}_secret"](name, **values):
            # Only read_secret raises exceptions sadly FIXME?
            raise CommandExecutionError(f"Failed to {verb} secret, see logs for details")
        ret["comment"] = f"The secret was {pp}"
    except SaltException as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}
    return ret


def absent(name, operation="delete"):
    """
    Ensure a secret is absent. This operates only on the most recent version
    for delete/destroy. Currently does not destroy/wipe a secret that has
    been made unreadable in some other way.

    name
        The path of the secret.

    operation
        The operation to perform to remove the secret. Only relevant for KV v2.
        Options are: ``delete`` (meaning: soft-delete), ``destroy`` (meaning delete unrecoverably)
        and ``wipe`` (forget about the secret completely). Defaults to ``delete``.
        KV v1 secrets are always wiped since the backend does not support versioning.
    """
    valid_ops = ("delete", "destroy", "wipe")
    if operation not in valid_ops:
        raise SaltInvocationError(f"Invalid operation '{operation}'. Valid: {', '.join(valid_ops)}")
    ret = {
        "name": name,
        "result": True,
        "comment": "The secret is already absent",
        "changes": {},
    }
    pp = "destroyed" if operation == "destroy" else operation + "d"
    try:
        try:
            __salt__["vault.read_secret"](name)
        except CommandExecutionError as err:
            if "VaultNotFoundError" not in str(err):
                raise
            return ret
        ret["changes"][pp] = name
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Would have {pp} the secret"
            return ret
        if not __salt__[f"vault.{operation}_secret"](name):
            # Only read_secret raises exceptions sadly FIXME?
            raise CommandExecutionError(f"Failed to {operation} secret, see logs for details")
        ret["comment"] = f"The secret has been {pp}"
    except SaltException as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}
    return ret



import os

import pytest

from tests.support.vault import vault_create_secret_id
from tests.support.vault import vault_delete_approle
from tests.support.vault import vault_disable_auth_method
from tests.support.vault import vault_enable_auth_method
from tests.support.vault import vault_get_role_id
from tests.support.vault import vault_write_approle

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides():
    return {
        "osenv": {"driver": "env"},
        "vault": {
            "auth": {
                "method": "token",
                "token": "sdb://osenv/VAULT_TOKEN",
            },
            "config_location": "local",
        },
    }


@pytest.fixture(scope="module")
def master_config_overrides():
    return {
        "osenv": {"driver": "env"},
        "vault": {
            "auth": {
                "method": "approle",
                "role_id": "sdb://osenv/VAULT_ROLEID",
                "secret_id": "sdb://osenv/VAULT_SECRETID",
            },
            "config_location": "local",
        },
    }


@pytest.fixture(scope="module")
def approle_configured():
    vault_enable_auth_method("approle", ["-path=approle"])
    vault_write_approle("sdbrole")
    try:
        role_id = vault_get_role_id("sdbrole")
        secret_id = vault_create_secret_id("sdbrole")
        yield {"role_id": role_id, "secret_id": secret_id}
    finally:
        vault_delete_approle("sdbrole")
        vault_disable_auth_method("approle")


@pytest.fixture(autouse=True)
def auth_in_env(approle_configured):
    prev, os.environ["VAULT_TOKEN"], os.environ["VAULT_ROLEID"], os.environ["VAULT_SECRETID"] = (
        os.environ.get("VAULT_TOKEN"),
        "testsecret",
        approle_configured["role_id"],
        approle_configured["secret_id"],
    )
    try:
        yield
    finally:
        if prev is not None:
            os.environ["VAULT_TOKEN"] = prev
        else:
            os.environ.pop("VAULT_TOKEN")
        os.environ.pop("VAULT_ROLEID")
        os.environ.pop("VAULT_SECRETID")


def test_sdb_in_config_token(salt_call_cli):
    res = salt_call_cli.run("vault.query", "GET", "auth/token/lookup-self")
    assert res.returncode == 0
    assert res.data["data"]["id"] == "testsecret"


def test_sdb_in_config_role_id_secret_id(salt_run_cli):
    res = salt_run_cli.run("vault.auth_info")
    assert res.returncode == 0
    assert res.data["token"]["meta"]["role_name"] == "sdbrole"



from unittest.mock import Mock

import pytest

from saltext.vault.modules import vault_db as vault_db_exe
from saltext.vault.states import vault_db


@pytest.fixture
def _conns():
    return {}


@pytest.fixture
def _roles():
    return {}


@pytest.fixture
def delete_connection_mock(_conns):
    def _del(name, **kwargs):  # pylint: disable=unused-argument
        _conns.pop(name, None)
        return True

    return Mock(spec=vault_db_exe.delete_connection, side_effect=_del)


@pytest.fixture
def delete_role_mock(_roles):
    def _del(name, **kwargs):  # pylint: disable=unused-argument
        _roles.pop(name, None)
        return True

    return Mock(spec=vault_db_exe.delete_role, side_effect=_del)


@pytest.fixture
def fetch_connection_mock(_conns):
    return Mock(spec=vault_db_exe.fetch_connection, side_effect=lambda x, **kwargs: _conns.get(x))


@pytest.fixture
def fetch_role_mock(_roles):
    return Mock(spec=vault_db_exe.fetch_role, side_effect=lambda x, **kwargs: _roles.get(x))


@pytest.fixture
def rotate_root_mock():
    return Mock(spec=vault_db_exe.rotate_root, return_value=True)


@pytest.fixture
def write_connection_mock(_conns):
    def _write(
        name,
        plugin,
        version="",
        verify=True,
        rotate=True,
        allowed_roles=None,
        root_rotation_statements=None,
        password_policy=None,
        mount="database",
        **kwargs,
    ):  # pylint: disable=unused-argument
        kwargs.pop("password", None)  # password is obviously never returned
        data = {
            "plugin_name": f"{plugin}-database-plugin",
            "plugin_version": version,
            "verify_connection": verify,
            "allowed_roles": allowed_roles or [],
            "root_credentials_rotate_statements": root_rotation_statements or [],
            "password_policy": password_policy or "",
            "connection_details": kwargs,
        }
        _conns[name] = data

    return Mock(spec=vault_db_exe.write_connection, side_effect=_write)


@pytest.fixture
def write_role_mock(_roles):
    def _write(
        name, connection, creation_statements, mount="database", **kwargs
    ):  # pylint: disable=unused-argument
        data = {
            "db_name": connection,
            "creation_statements": creation_statements,
            "default_ttl": kwargs.get("default_ttl") or 0,
            "max_ttl": kwargs.get("max_ttl") or 0,
            "renew_statements": kwargs.get("renew_statements") or [],
            "revocation_statements": kwargs.get("revocation_statements") or [],
            "rollback_statements": kwargs.get("rollback_statements") or [],
            # credential_type/_config are only exposed if the DB plugin supports it
        }
        _roles[name] = data

    return Mock(spec=vault_db_exe.write_role, side_effect=_write)


@pytest.fixture
def write_static_role_mock(_roles):
    def _write(
        name,
        connection,
        username,
        rotation_period,
        rotation_statements=None,
        mount="database",
        **kwargs,
    ):  # pylint: disable=unused-argument
        data = {
            "db_name": connection,
            "username": username,
            "rotation_period": rotation_period,
            "rotation_statements": rotation_statements or [],
            # credential_type/_config are only exposed if the DB plugin supports it
        }
        _roles[name] = data

    return Mock(spec=vault_db_exe.write_static_role, side_effect=_write)


@pytest.fixture
def configure_loader_modules(
    delete_connection_mock,
    delete_role_mock,
    fetch_connection_mock,
    fetch_role_mock,
    rotate_root_mock,
    write_connection_mock,
    write_role_mock,
    write_static_role_mock,
    testmode,
):
    return {
        vault_db: {
            "__opts__": {"test": testmode},
            "__grains__": {"id": "test-minion"},
            "__salt__": {
                "vault_db.delete_connection": delete_connection_mock,
                "vault_db.delete_role": delete_role_mock,
                "vault_db.fetch_connection": fetch_connection_mock,
                "vault_db.fetch_role": fetch_role_mock,
                "vault_db.write_connection": write_connection_mock,
                "vault_db.write_role": write_role_mock,
                "vault_db.write_static_role": write_static_role_mock,
                "vault_db.rotate_root": rotate_root_mock,
            },
        }
    }


@pytest.fixture(params=(False, True), autouse=True)
def testmode(request):
    return request.param


def test_conn_present(testmode, write_connection_mock):
    ret = vault_db.connection_present("conn", "custom")
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert "created" in ret["changes"]
    assert (write_connection_mock.call_args is None) is testmode
    assert "created" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode


def test_conn_already_present(write_connection_mock):
    write_connection_mock("conn", "custom")
    ret = vault_db.connection_present("conn", "custom")
    assert ret["result"] is True
    assert not ret["changes"]
    assert "as specified" in ret["comment"]


@pytest.mark.parametrize(
    "kwargs,param",
    (
        ({"version": "1.2.3"}, "plugin_version"),
        ({"allowed_roles": ["*"]}, None),
        ({"root_rotation_statements": ["rotate"]}, "root_credentials_rotate_statements"),
        ({"password_policy": "foo"}, None),
    ),
)
def test_conn_changes(testmode, write_connection_mock, kwargs, param, _conns):
    write_connection_mock("conn", "custom")
    param = param or next(iter(kwargs))
    expected_changes = {"old": _conns["conn"][param], "new": kwargs[next(iter(kwargs))]}
    ret = vault_db.connection_present("conn", "custom", **kwargs)
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert param in ret["changes"]
    assert ret["changes"][param] == expected_changes
    assert (write_connection_mock.call_count == 1) is testmode
    assert "updated" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode


def test_conn_no_password_changes(write_connection_mock, _conns):
    write_connection_mock("conn", "custom", username="foo", password="bar")
    ret = vault_db.connection_present("conn", "custom", username="foo", password="baz")
    assert ret["result"] is True
    assert not ret["changes"]


def test_conn_detail_changes(testmode, write_connection_mock, _conns):
    write_connection_mock("conn", "custom", username="foo", password="bar")
    ret = vault_db.connection_present("conn", "custom", username="bar", password="baz")
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert "username" in ret["changes"]
    assert ret["changes"]["username"] == {"old": "foo", "new": "bar"}
    assert (write_connection_mock.call_count == 1) is testmode
    assert "updated" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode
    if not testmode:
        assert "username" in write_connection_mock.call_args[1]
        # for an existing connection, the password should never be updated
        assert "password" not in write_connection_mock.call_args[1]


def test_conn_statements_strip(write_connection_mock):
    write_connection_mock("conn", "custom", root_rotation_statements=["foo"])
    ret = vault_db.connection_present("conn", "custom", root_rotation_statements=["foo\n"])
    assert ret["result"] is True
    assert not ret["changes"]


def test_conn_plugin_change_err(write_connection_mock):
    write_connection_mock("conn", "custom")
    ret = vault_db.connection_present("conn", "custom2")
    assert ret["result"] is False
    assert not ret["changes"]
    assert "Cannot change plugin type without deleting" in ret["comment"]


def test_conn_plugin_change_force(testmode, write_connection_mock, delete_connection_mock):
    write_connection_mock("conn", "custom")
    ret = vault_db.connection_present("conn", "custom2", force=True)
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert "deleted_for_plugin_change" in ret["changes"]
    assert (write_connection_mock.call_count == 1) is testmode
    assert "created" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode
    assert (delete_connection_mock.call_count == 0) is testmode


@pytest.mark.parametrize("testmode", (False,), indirect=True)
@pytest.mark.parametrize("present", (False, True))
def test_conn_verification(present, write_connection_mock):
    if present:
        write_connection_mock("conn", "custom")
    write_connection_mock.side_effect = None
    ret = vault_db.connection_present("conn", "custom", allowed_roles=["*"])
    assert ret["result"] is False
    if present:
        assert "reported parameters do not match" in ret["comment"]
    else:
        assert "but it is still reported as absent" in ret["comment"]


def test_conn_absent(testmode, write_connection_mock, delete_connection_mock):
    write_connection_mock("conn", "custom")
    ret = vault_db.connection_absent("conn")
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert "deleted" in ret["changes"]
    assert "deleted" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode
    assert (delete_connection_mock.call_count == 0) is testmode


def test_conn_already_absent(delete_connection_mock):
    ret = vault_db.connection_absent("conn", "custom")
    assert ret["result"] is True
    assert not ret["changes"]
    assert "already absent" in ret["comment"]
    delete_connection_mock.assert_not_called()


@pytest.mark.parametrize("testmode", (False,), indirect=True)
def test_conn_absent_verification(write_connection_mock, delete_connection_mock):
    write_connection_mock("conn", "custom")
    delete_connection_mock.side_effect = None
    ret = vault_db.connection_absent("conn", "custom")
    assert ret["result"] is False
    assert "but it is still reported as present" in ret["comment"]


def test_role_present(testmode, write_role_mock):
    ret = vault_db.role_present("role", "conn", [])
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert "created" in ret["changes"]
    assert (write_role_mock.call_args is None) is testmode
    assert "created" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode


def test_role_already_present(write_role_mock):
    write_role_mock("role", "conn", [])
    ret = vault_db.role_present("role", "conn", [])
    assert ret["result"] is True
    assert not ret["changes"]
    assert "as specified" in ret["comment"]


@pytest.mark.parametrize(
    "kwargs,param",
    (
        ({"connection": "conn2"}, "db_name"),
        ({"creation_statements": ["foo"]}, None),
        ({"revocation_statements": ["revoke"]}, None),
        ({"rollback_statements": ["back"]}, None),
        ({"renew_statements": ["bling"]}, None),
        ({"default_ttl": 42}, None),
        ({"max_ttl": 1337}, None),
    ),
)
def test_role_changes(testmode, write_role_mock, kwargs, param, _roles):
    write_role_mock("role", "conn", [])
    kwargs = kwargs.copy()
    param = param or next(iter(kwargs))
    expected_changes = {"old": _roles["role"][param], "new": kwargs[next(iter(kwargs))]}
    conn = kwargs.pop("connection", "conn")
    creation_statements = kwargs.pop("creation_statements", [])
    ret = vault_db.role_present("role", conn, creation_statements, **kwargs)
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert param in ret["changes"]
    assert ret["changes"][param] == expected_changes
    assert (write_role_mock.call_count == 1) is testmode
    assert "updated" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode


def test_role_changes_strip(write_role_mock):
    write_role_mock(
        "role",
        "conn",
        creation_statements=["foo"],
        revocation_statements=["foo"],
        rollback_statements=["foo"],
        renew_statements=["foo"],
    )
    ret = vault_db.role_present(
        "role",
        "conn",
        creation_statements=["foo\n"],
        revocation_statements=["foo\n"],
        rollback_statements=["foo\n"],
        renew_statements=["foo\n"],
    )
    assert ret["result"] is True
    assert not ret["changes"]


def test_role_statements_as_strings(write_role_mock):
    write_role_mock(
        "role",
        "conn",
        creation_statements=["foo"],
        revocation_statements=["foo"],
        rollback_statements=["foo"],
        renew_statements=["foo"],
    )
    ret = vault_db.role_present(
        "role",
        "conn",
        creation_statements="foo",
        revocation_statements="foo",
        rollback_statements="foo",
        renew_statements="foo",
    )
    assert ret["result"] is True
    assert not ret["changes"]


@pytest.mark.parametrize("testmode", (False,), indirect=True)
@pytest.mark.parametrize("present", (False, True))
def test_role_verification(present, write_role_mock):
    if present:
        write_role_mock("role", "conn", [])
    write_role_mock.side_effect = None
    ret = vault_db.role_present("role", "conn", ["foo"])
    assert ret["result"] is False
    if present:
        assert "reported parameters do not match" in ret["comment"]
    else:
        assert "but it is still reported as absent" in ret["comment"]


def test_role_absent(testmode, write_role_mock, delete_role_mock):
    write_role_mock("role", "conn", [])
    ret = vault_db.role_absent("role")
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert "deleted" in ret["changes"]
    assert "deleted" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode
    assert (delete_role_mock.call_count == 0) is testmode


def test_role_already_absent(delete_role_mock):
    ret = vault_db.role_absent("role")
    assert ret["result"] is True
    assert not ret["changes"]
    assert "already absent" in ret["comment"]
    delete_role_mock.assert_not_called()


@pytest.mark.parametrize("testmode", (False,), indirect=True)
def test_role_absent_verification(write_role_mock, delete_role_mock):
    write_role_mock("role", "conn", [])
    delete_role_mock.side_effect = None
    ret = vault_db.role_absent("role")
    assert ret["result"] is False
    assert "but it is still reported as present" in ret["comment"]


def test_static_role_present(testmode, write_static_role_mock):
    ret = vault_db.static_role_present("role", "conn", "user", 42)
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert "created" in ret["changes"]
    assert (write_static_role_mock.call_args is None) is testmode
    assert "created" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode


def test_static_role_already_present(write_static_role_mock):
    write_static_role_mock("role", "conn", "user", 42)
    ret = vault_db.static_role_present("role", "conn", "user", 42)
    assert ret["result"] is True
    assert not ret["changes"]
    assert "as specified" in ret["comment"]


@pytest.mark.parametrize(
    "kwargs,param",
    (
        ({"connection": "conn2"}, "db_name"),
        ({"username": "bar"}, None),
        ({"rotation_period": 43}, None),
    ),
)
def test_static_role_changes(testmode, write_static_role_mock, kwargs, param, _roles):
    write_static_role_mock("role", "conn", "user", 42)
    kwargs = kwargs.copy()
    param = param or next(iter(kwargs))
    expected_changes = {"old": _roles["role"][param], "new": kwargs[next(iter(kwargs))]}
    conn = kwargs.pop("connection", "conn")
    username = kwargs.pop("username", "user")
    rotation_period = kwargs.pop("rotation_period", 42)
    ret = vault_db.static_role_present("role", conn, username, rotation_period, **kwargs)
    assert ret["result"] is not False
    assert (ret["result"] is None) is testmode
    assert ret["changes"]
    assert param in ret["changes"]
    assert ret["changes"][param] == expected_changes
    assert (write_static_role_mock.call_count == 1) is testmode
    assert "updated" in ret["comment"]
    assert ("would have been" in ret["comment"]) is testmode


def test_static_role_changes_strip(write_static_role_mock):
    write_static_role_mock("role", "conn", "user", 42, rotation_statements=["foo"])
    ret = vault_db.static_role_present("role", "conn", "user", 42, rotation_statements=["foo\n"])
    assert ret["result"] is True
    assert not ret["changes"]


def test_static_role_statements_as_strings(write_static_role_mock):
    write_static_role_mock("role", "conn", "user", 42, rotation_statements=["foo"])
    ret = vault_db.static_role_present("role", "conn", "user", 42, rotation_statements="foo")
    assert ret["result"] is True
    assert not ret["changes"]


@pytest.mark.parametrize("testmode", (False,), indirect=True)
@pytest.mark.parametrize("present", (False, True))
def test_static_role_verification(present, write_static_role_mock):
    if present:
        write_static_role_mock("role", "conn", "user", 42)
    write_static_role_mock.side_effect = None
    ret = vault_db.static_role_present("role", "conn", "user", 43)
    assert ret["result"] is False
    if present:
        assert "reported parameters do not match" in ret["comment"]
    else:
        assert "but it is still reported as absent" in ret["comment"]



"""
Manage the Vault (or OpenBao) SSH secret engine, request SSH credentials
and certificates.

.. versionadded:: 1.2.0

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import logging
from pathlib import Path

import salt.utils.json
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

from saltext.vault.utils import vault
from saltext.vault.utils.vault.helpers import deserialize_csl

__virtualname__ = "vault_ssh"

log = logging.getLogger(__name__)


def __virtual__():
    return __virtualname__


def read_role(name, mount="ssh"):
    """
    Reads an existing SSH role.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.read_role sre

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/roles/<name>" {
            capabilities = ["read"]
        }

    name
        The name of the SSH role.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    try:
        return vault.query("GET", f"{mount}/roles/{name}", __opts__, __context__)["data"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def write_role_otp(
    name,
    default_user,
    cidr_list=None,
    allowed_users=None,
    exclude_cidr_list=None,
    port=None,
    mount="ssh",
):
    """
    Create/update an SSH role (OTP type).

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.write_role_otp sre sre-user '["1.0.0.0/24"]'

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/roles/<name>" {
            capabilities = ["create", "update"]
        }

    name
        The name of the SSH role.

    default_user
        The default username for which a credential will be generated.
        Required.

    cidr_list
        List of CIDR blocks to which the role is applicable.
        Required, unless the role is registered as a zero-address role.

    allowed_users
        List of usernames the client can request under this role.
        By default, **any usernames are allowed** (``*``).
        The ``default_user`` will always be allowed.

    exclude_cidr_list
        List of CIDR blocks not accepted by the role.

    port
        Specifies the port number for SSH connections, which will be returned to
        OTP clients as an informative value. Defaults to ``22``.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    payload = {
        "default_user": default_user,
        "cidr_list": cidr_list,
        "allowed_users": allowed_users,
        "exclude_cidr_list": exclude_cidr_list,
        "port": int(port) if port else None,
    }
    for param in ("allowed_users", "cidr_list", "exclude_cidr_list"):
        payload[param] = deserialize_csl(payload[param])
    return _write_role(name, "otp", mount=mount, **payload)


def write_role_ca(
    name,
    default_user="",
    default_user_template=None,
    allowed_users=None,
    allowed_users_template=None,
    allowed_domains=None,
    allowed_domains_template=None,
    ttl=None,
    max_ttl=None,
    allowed_critical_options=None,
    allowed_extensions=None,
    default_critical_options=None,
    default_extensions=None,
    default_extensions_template=None,
    allow_user_certificates=None,
    allow_host_certificates=None,
    allow_bare_domains=None,
    allow_subdomains=None,
    allow_user_key_ids=None,
    key_id_format=None,
    allowed_user_key_lengths=None,
    algorithm_signer=None,
    not_before_duration=None,
    mount="ssh",
):
    """
    Create/update an SSH role (CA type).

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.write_role_ca sre allowed_users=[sre-user] allow_user_certificates=true

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/roles/<name>" {
            capabilities = ["create", "update"]
        }

    name
        The name of the SSH role.

    default_user
        The default username for which a credential will be generated.
        When ``default_user_template`` is true, this can contain an identity
        template with any prefix or suffix, like ``ssh-{{identity.entity.id}}-user``.
        If you wish this to be a valid principal, it must also be in ``allowed_users``.

    default_user_template
        Allow ``default_users`` to be specified using identity template values.
        A non-templated user is also permitted. Defaults to false.

    allowed_users
        List of usernames the client can request under this role.
        By default, none are allowed. Set this to ``*`` to allow any usernames.
        If ``allowed_users_template`` is true, this list can contain an
        identity template with any prefix or suffix. The ``default_user``
        will always be allowed.

    allowed_users_template
        Allow ``allowed_users`` to be specified using identity template values.
        Non-templated users are also permitted. Defaults to false.

    allowed_domains
        List of domains for which a client can request a host certificate.
        ``*`` allows any domain. See also ``allow_bare_domains`` and ``allow_subdomains``.

    allowed_domains_template
        Allow ``allowed_domains`` to be specified using identity template values.
        Non-templated domains are also permitted. Defaults to false.

    ttl
        Specifies the Time To Live value provided as a string duration with
        time suffix. Hour is the largest suffix. If unset, uses the system
        default value or the value of ``max_ttl``, whichever is shorter

    max_ttl
        Specifies the maximum Time To Live provided as a string duration with
        time suffix. Hour is the largest suffix. If unset, defaults to the
        system maximum lease TTL.

    allowed_critical_options
        List of critical options that certificates can carry when signed.
        If unset (default), allows any option.

    allowed_extensions
        List of extensions that certificates can carry when signed.
        If unset (default), will always take the extensions
        from ``default_extensions`` only. If set to ``*``, will allow
        any extension to be set.
        For the list of extensions, take a look at the sshd manual's
        AUTHORIZED_KEYS FILE FORMAT section. You should add a ``permit-``
        before the name of extension to allow it.

    default_critical_options
        Map of critical options to their values certificates should carry
        if none are provided when signing.

    default_extensions
        Map of extensions to their values certificates should carry
        if none are provided when signing or allowed_extensions is unset.

    default_extensions_template
        Allow ``default_extensions`` to be specified using identity template values.
        Non-templated values are also permitted. Defaults to false.

    allow_user_certificates
        Allow certificates to be signed for ``user`` use. Defaults to false.

    allow_host_certificates
        Allow certificates to be signed for ``host`` use. Defaults to false.

    allow_bare_domains
        Allow host certificates to be signed for the base domains listed in
        ``allowed_domains``. This is a separate option as in some cases this
        can be considered a security threat. Defaults to false.

    allow_subdomains
        Allow host certificates to be signed for subdomains of the base domains
        listed in ``allowed_domains``. Defaults to false.

    allow_user_key_ids
        Allow users to override the key ID for a certificate. When false (default),
        the key ID will always be the token display name.
        The key ID is logged by the SSH server and can be useful for auditing.

    key_id_format
        Specifies a custom format for the key ID of a signed certificate.
        See `key_id_format <https://developer.hashicorp.com/vault/api-docs/secret/ssh#key_id_format>`_
        for available template values.

    allowed_user_key_lengths
        Map of ssh key types to allowed sizes when signing with the CA type.
        Values can be a list of multiple sizes.
        Keys can both be OpenSSH-style key identifiers and short names
        (``rsa``, ``ecdsa``, ``dsa``, or ``ed25519``). If an algorithm has
        a fixed key size, values are ignored.

    algorithm_signer
        **RSA** algorithm to sign keys with. Valid: ``ssh-rsa``, ``rsa-sha2-256``,
        ``rsa-sha2-512``, or ``default`` (which is the default). Ignored
        when not signing with an RSA key.

    not_before_duration
        Specifies the duration by which to backdate the ``ValidAfter`` property.
        Defaults to ``30s``.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    if not (allow_user_certificates or allow_host_certificates):
        raise SaltInvocationError(
            "Either allow_user_certificates or allow_host_certificates must be true"
        )

    payload = {
        "default_user": default_user,
        "default_user_template": default_user_template,
        "allowed_users": allowed_users,
        "allowed_users_template": allowed_users_template,
        "allowed_domains": allowed_domains,
        "allowed_domains_template": allowed_domains_template,
        "ttl": ttl,
        "max_ttl": max_ttl,
        "allowed_critical_options": allowed_critical_options,
        "allowed_extensions": allowed_extensions,
        "default_critical_options": default_critical_options,
        "default_extensions": default_extensions,
        "default_extensions_template": default_extensions_template,
        "allow_user_certificates": allow_user_certificates,
        "allow_host_certificates": allow_host_certificates,
        "allow_bare_domains": allow_bare_domains,
        "allow_subdomains": allow_subdomains,
        "allow_user_key_ids": allow_user_key_ids,
        "key_id_format": key_id_format,
        "allowed_user_key_lengths": allowed_user_key_lengths,
        "algorithm_signer": algorithm_signer,
        "not_before_duration": not_before_duration,
    }
    for param in (
        "allowed_users",
        "allowed_critical_options",
        "allowed_domains",
        "allowed_extensions",
    ):
        payload[param] = deserialize_csl(payload[param])

    for param in ("default_critical_options", "default_extensions", "allowed_user_key_lengths"):
        if payload[param] and not isinstance(payload[param], dict):
            raise SaltInvocationError(f"Parameter '{param}' must be specified as a mapping")
    return _write_role(name, "ca", mount=mount, **payload)


def _write_role(name, key_type, mount="ssh", **kwargs):
    endpoint = f"{mount}/roles/{name}"
    kwargs["key_type"] = key_type

    payload = {}
    for param, val in kwargs.items():
        if isinstance(val, list):
            payload[param] = ",".join(val)
        elif val is not None:
            payload[param] = val
    try:
        return vault.query(
            "POST", endpoint, __opts__, __context__, payload=payload, safe_to_retry=True
        )
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def delete_role(name, mount="ssh"):
    """
    Deletes an existing SSH role.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.delete_role sre

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/roles/<name>" {
            capabilities = ["delete"]
        }

    name
        The name of the SSH role.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    try:
        return vault.query("DELETE", f"{mount}/roles/{name}", __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def list_roles(mount="ssh"):
    """
    Lists existing SSH roles.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.list_roles

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/roles" {
            capabilities = ["list"]
        }

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    try:
        res = vault.query("LIST", f"{mount}/roles", __opts__, __context__)["data"]
    except vault.VaultNotFoundError:
        return {}
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err

    keys = res["key_info"]
    for key in res["keys"]:
        if key not in keys:
            keys[key] = {}
    return keys


def list_roles_ip(address, mount="ssh"):
    """
    Lists existing SSH roles associated with a given IP address.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.list_roles_ip 10.1.0.1

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/lookup" {
            capabilities = ["create", "update"]
        }

    address
        The IP address to list roles for.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    endpoint = f"{mount}/lookup"
    payload = {"ip": address}
    try:
        return vault.query(
            "POST", endpoint, __opts__, __context__, payload=payload, safe_to_retry=True
        )["data"]["roles"]
    except vault.VaultInvocationError as err:
        if "Missing roles" not in str(err):
            raise
        return []
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def list_roles_zeroaddr(mount="ssh"):
    """
    Return the list of configured zero-address roles. These are roles
    that are allowed to request credentials for any IP address.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.list_roles_zeroaddr

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/config/zeroaddress" {
            capabilities = ["read"]
        }

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    try:
        return vault.query("GET", f"{mount}/config/zeroaddress", __opts__, __context__)["data"][
            "roles"
        ]
    except vault.VaultNotFoundError:
        return []
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def write_zeroaddr_roles(roles, mount="ssh"):
    """
    Write the list of configured zero-address roles. These are roles
    that are allowed to request credentials for any IP address.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.write_roles_zeroaddr '[super, admin]'

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/config/zeroaddress" {
            capabilities = ["create", "update"]
        }

    roles
        The list of role names that should be marked as zero address roles.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    endpoint = f"{mount}/config/zeroaddress"
    payload = {"roles": roles}
    try:
        return vault.query(
            "POST", endpoint, __opts__, __context__, payload=payload, safe_to_retry=True
        )
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def delete_zeroaddr_roles(mount="ssh"):
    """
    Delete the list of configured zero-address roles. These are roles
    that are allowed to request credentials for any IP address.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.delete_roles_zeroaddr

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/config/zeroaddress" {
            capabilities = ["delete"]
        }

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    try:
        return vault.query("DELETE", f"{mount}/config/zeroaddress", __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def get_creds(name, address, username="", mount="ssh"):
    """
    Generate credentials for a specific IP (and username) using an existing role.
    Returns a mapping with ``ip``, ``key``, ``key_type``, ``port`` and ``username``.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.get_creds sre 10.1.0.1
        salt '*' vault_ssh.get_creds sre 10.1.0.1 bob

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/creds/<role_name>" {
            capabilities = ["create", "update"]
        }

    name
        The name of the role.

    address
        The IP address of the host to generate credentials for.

    username
        The username on the remote host to generate credentials for.
        If empty, the default username of the role will be used.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    endpoint = f"{mount}/creds/{name}"
    payload = {"ip": address, "username": username}
    # TODO: cache lease!
    try:
        return vault.query("POST", endpoint, __opts__, __context__, payload=payload)["data"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def create_ca(
    private_key=None,
    public_key=None,
    key_type="ssh-rsa",
    key_bits=0,
    mount="ssh",
):
    """
    Create a CA to be used for certificate authentication.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.create_ca

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/config/ca" {
            capabilities = ["create", "update"]
        }

    private_key
        The private key part of the SSH CA key pair. Can be a file
        local to the minion or a PEM-encoded string.
        If this or ``public_key`` is unspecified, will generate a pair
        on the Vault server.

    public_key
        The public key part of the SSH CA key pair. Can be a file
        local to the minion or a PEM-encoded string.
        If this or ``public_key`` is unspecified, will generate a pair
        on the Vault server.

    key_type
        The desired key type for the generated SSH CA key when generating
        on the Vault server. Valid: ``ssh-rsa`` (default), ``sha2-nistp256``,
        ``ecdsa-sha2-nistp384``, ``ecdsa-sha2-nistp521``, or ``ssh-ed25519``.
        Can also specify an algorithm: ``rsa``, ``ec``, or ``ed25519``.

    key_bits
        The desired key bits for the generated SSH CA key when generating
        on the Vault server. Only used for variable length keys (e.g. ``ssh-rsa``)
        or when ``ec`` was specified as ``key_type``, in which case this
        selects the NIST P-curve: ``256``, ``384``, ``521``.
        0 (default) will select 4096 bits for RSA or NIST P-256 for EC.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    endpoint = f"{mount}/config/ca"

    if private_key and public_key:
        payload = {
            "generate_signing_key": False,
            "private_key": _get_file_or_data(private_key),
            "public_key": _get_file_or_data(public_key),
        }
    else:
        payload = {
            "generate_signing_key": True,
            "key_type": key_type,
            "key_bits": key_bits,
        }

    try:
        res = vault.query("POST", endpoint, __opts__, __context__, payload=payload)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err
    try:
        return res["data"]["public_key"]
    except TypeError:
        # when keys have been imported, we just receive a boolean
        return read_ca(mount=mount)


def destroy_ca(mount="ssh"):
    """
    Destroy an existing CA on the mount.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.destroy_ca

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/config/ca" {
            capabilities = ["delete"]
        }

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    try:
        return vault.query("DELETE", f"{mount}/config/ca", __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def read_ca(mount="ssh"):
    """
    Read the public key for an existing CA on the mount.
    This defaults to reading from the authenticated endpoint, but falls
    back to the unauthenticated one.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.read_ca

    Required policy: None for the unauthenticated endpoint or

    .. code-block:: vaultpolicy

        path "<mount>/config/ca" {
            capabilities = ["read"]
        }

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    try:
        return vault.query("GET", f"{mount}/config/ca", __opts__, __context__)["data"]["public_key"]
    except vault.VaultPermissionDeniedError:
        log.info("Permission denied for the authenticated endpoint, trying unauthenticated one")
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err
    try:
        res = vault.query_raw("GET", f"{mount}/public_key", __opts__, __context__, is_unauthd=True)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err
    if res.status_code == 200:
        return res.text
    res.raise_for_status()
    raise CommandExecutionError(
        f"Internal error, this should not have been hit. Response ({res.status_code}): {res.text}"
    )


def sign_key(
    name,
    public_key,
    ttl=None,
    valid_principals=None,
    cert_type="user",
    key_id=None,
    critical_options=None,
    extensions=None,
    mount="ssh",
):
    """
    Sign an SSH public key under an existing role on the mount.
    Returns a mapping with ``serial_number`` and ``signed_key``.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.sign_key sre $HOME/.ssh/id_me.pub

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/sign/<role_name>" {
            capabilities = ["create", "update"]
        }

    name
        The name of the SSH role.

    public_key
        The SSH public key that should be signed. Can be a file local to
        the minion or a PEM-encoded string.

    ttl
        Request a specific time to live for the certificate, limited by the
        role's TTL. If unspecified, will default to the role's TTL or system
        values.

    valid_principals
        List of usernames/hostnames the certificate should be signed for.

    cert_type
        The type of certificate to issue, either ``user`` or ``host``. Defaults
        to ``user``.

    key_id
        The key ID the created certificate should have. If unspecified, the display
        name of the creating token will be used.

    critical_options
        A map of critical options the certificate should carry.

    extensions
        A map of extensions the certificate should carry.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    endpoint = f"{mount}/sign/{name}"
    payload = {"public_key": _get_file_or_data(public_key), "cert_type": cert_type}

    if ttl is not None:
        payload["ttl"] = ttl
    if key_id is not None:
        payload["key_id"] = key_id

    if valid_principals is not None:
        if isinstance(valid_principals, list):
            valid_principals = ",".join(valid_principals)
        payload["valid_principals"] = valid_principals

    for param, val in [
        ("critical_options", critical_options),
        ("extensions", extensions),
    ]:
        if val is not None:
            payload[param] = {
                k: salt.utils.json.dumps(v) if isinstance(v, (dict, list)) else v
                for k, v in val.items()
            }

    try:
        return vault.query("POST", endpoint, __opts__, __context__, payload=payload)["data"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def generate_key_cert(
    name,
    key_type="ssh-rsa",
    key_bits=0,
    ttl=None,
    valid_principals=None,
    cert_type="user",
    key_id=None,
    critical_options=None,
    extensions=None,
    mount="ssh",
):
    """
    Generate an SSH private key and accompanying signed certificate.
    Returns a mapping with keys ``private_key``, ``private_key_type``,
    ``serial_number``, ``signed_key``.

    CLI Example:

    .. code-block:: bash

        salt '*' vault_ssh.generate_key_cert sre

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/sign/<role_name>" {
            capabilities = ["create", "update"]
        }

    name
        The name of the SSH role.

    key_type
        The desired key type for the generated SSH CA key.
        Valid: ``ssh-rsa`` (default), ``sha2-nistp256``,
        ``ecdsa-sha2-nistp384``, ``ecdsa-sha2-nistp521``, or ``ssh-ed25519``.
        Can also specify an algorithm: ``rsa``, ``ec``, or ``ed25519``.

    key_bits
        The desired key bits for the generated SSH CA key.
        Only used for variable length keys (e.g. ``ssh-rsa``)
        or when ``ec`` was specified as ``key_type``, in which case this
        selects the NIST P-curve: ``256``, ``384``, ``521``.
        0 (default) will select 4096 bits for RSA or NIST P-256 for EC.

    ttl
        Request a specific time to live for the certificate, limited by the
        role's TTL. If unspecified, will default to the role's TTL or system
        values.

    valid_principals
        List of usernames/hostnames the certificate should be signed for.

    cert_type
        The type of certificate to issue, either ``user`` or ``host``. Defaults
        to ``user``.

    key_id
        The key ID the created certificate should have. If unspecified, the display
        name of the creating token will be used.

    critical_options
        A map of critical options the certificate should carry.

    extensions
        A map of extensions the certificate should carry.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    endpoint = f"{mount}/issue/{name}"
    payload = {"key_type": key_type, "cert_type": cert_type, "key_bits": key_bits}

    if ttl is not None:
        payload["ttl"] = ttl
    if key_id is not None:
        payload["key_id"] = key_id

    if valid_principals is not None:
        if isinstance(valid_principals, list):
            valid_principals = ",".join(valid_principals)
        payload["valid_principals"] = valid_principals

    for param, val in [
        ("critical_options", critical_options),
        ("extensions", extensions),
    ]:
        if val is not None:
            payload[param] = {
                k: salt.utils.json.dumps(v) if isinstance(v, (dict, list)) else v
                for k, v in val.items()
            }

    try:
        return vault.query("POST", endpoint, __opts__, __context__, payload=payload)["data"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def _get_file_or_data(data):
    """
    Try to load a string as a file, otherwise return the string
    """
    try:
        # Check if the data can be interpreted as a Path at all
        Path(data)
    except TypeError:
        return data
    try:
        if __salt__["file.file_exists"](data):
            return __salt__["file.read"](data)
    except (OSError, TypeError, ValueError):
        pass
    return data



import logging
import subprocess

import pytest
import salt.utils.platform
import salt.version

pytest.importorskip("docker")

log = logging.getLogger(__name__)
salt_version = int(salt.version.__version__.split(".")[0])

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.skipif(salt_version < 3007, reason="Master cluster requires Salt 3007+"),
    pytest.mark.usefixtures("container", "vault_testing_values"),
]


@pytest.fixture(scope="module")
def vault_master_config(vault_port):
    return {
        "open_mode": True,
        "ext_pillar": [{"vault": "secret/path/foo"}],
        "peer_run": {
            ".*": [
                "vault.get_config",
                "vault.generate_new_token",
            ],
        },
        "vault": {
            "auth": {"token": "testsecret"},
            "cache": {
                "backend": "file",
            },
            "issue": {
                "type": "token",
                "token": {
                    "params": {
                        "num_uses": 0,
                    }
                },
            },
            "policies": {
                "assign": [
                    "salt_minion",
                    "salt_minion_{minion}",
                    "salt_role_{pillar[roles]}",
                ],
                "cache_time": 0,
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
        "minion_data_cache": True,
    }


@pytest.fixture(scope="module")
def cluster_shared_path(tmp_path_factory):
    return tmp_path_factory.mktemp("cluster")


@pytest.fixture(scope="module")
def cluster_pki_path(cluster_shared_path):
    path = cluster_shared_path / "pki"
    path.mkdir()
    (path / "peers").mkdir()
    return path


@pytest.fixture(scope="module")
def cluster_cache_path(cluster_shared_path):
    path = cluster_shared_path / "cache"
    path.mkdir()
    return path


@pytest.fixture(scope="module")
def cluster_master_1(salt_factories, cluster_pki_path, cluster_cache_path, vault_master_config):
    config_overrides = {
        "interface": "127.0.0.1",
        "cluster_id": "master_cluster",
        "cluster_peers": [
            "127.0.0.2",
            "127.0.0.3",
        ],
        "cluster_pki_dir": str(cluster_pki_path),
        "cache_dir": str(cluster_cache_path),
    }
    factory = salt_factories.salt_master_daemon(
        "127.0.0.1",
        defaults=vault_master_config,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture(scope="module")
def cluster_master_2(salt_factories, cluster_master_1, vault_master_config):
    if salt.utils.platform.is_darwin() or salt.utils.platform.is_freebsd():
        subprocess.check_output(["ifconfig", "lo0", "alias", "127.0.0.2", "up"])

    config_overrides = {
        "interface": "127.0.0.2",
        "cluster_id": "master_cluster",
        "cluster_peers": [
            "127.0.0.1",
            "127.0.0.3",
        ],
        "cluster_pki_dir": cluster_master_1.config["cluster_pki_dir"],
        "cache_dir": cluster_master_1.config["cache_dir"],
    }

    # Use the same ports for both masters, they are binding to different interfaces
    for key in (
        "ret_port",
        "publish_port",
    ):
        config_overrides[key] = cluster_master_1.config[key]
    factory = salt_factories.salt_master_daemon(
        "127.0.0.2",
        defaults=vault_master_config,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture(scope="module")
def cluster_minion_1(cluster_master_1, vault_master_config):
    port = cluster_master_1.config["ret_port"]
    addr = cluster_master_1.config["interface"]
    config_overrides = {
        "master": f"{addr}:{port}",
    }
    factory = cluster_master_1.salt_minion_daemon(
        "cluster-minion-1",
        defaults=vault_master_config,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture
def salt_call_cli(cluster_minion_1):
    return cluster_minion_1.salt_call_cli(timeout=120)


def test_minion_can_authenticate(salt_call_cli):
    ret = salt_call_cli.run("vault.read_secret", "secret/path/foo")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data.get("success") == "yeehaaw"


def test_minion_pillar_is_populated_as_expected(salt_call_cli):
    ret = salt_call_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data.get("success") == "yeehaaw"



"""
Vault Database helpers

.. versionadded:: 1.1.0
"""

from salt.utils.immutabletypes import freeze

PLUGINS = freeze(
    {
        "cassandra": {
            "name": "cassandra",
            "required": [
                "hosts",
                "username",
                "password",
            ],
        },
        "couchbase": {
            "name": "couchbase",
            "required": [
                "hosts",
                "username",
                "password",
            ],
        },
        "elasticsearch": {
            "name": "elasticsearch",
            "required": [
                "url",
                "username",
                "password",
            ],
        },
        "influxdb": {
            "name": "influxdb",
            "required": [
                "host",
                "username",
                "password",
            ],
        },
        "hanadb": {
            "name": "hana",
            "required": [
                "connection_url",
            ],
        },
        "mongodb": {
            "name": "mongodb",
            "required": [
                "connection_url",
            ],
        },
        "mongodb_atlas": {
            "name": "mongodbatlas",
            "required": [
                "public_key",
                "private_key",
                "project_id",
            ],
        },
        "mssql": {
            "name": "mssql",
            "required": [
                "connection_url",
            ],
        },
        "mysql": {
            "name": "mysql",
            "required": [
                "connection_url",
            ],
        },
        "oracle": {
            "name": "oracle",
            "required": [
                "connection_url",
            ],
        },
        "postgresql": {
            "name": "postgresql",
            "required": [
                "connection_url",
            ],
        },
        "redis": {
            "name": "redis",
            "required": [
                "host",
                "port",
                "username",
                "password",
            ],
        },
        "redis_elasticache": {
            "name": "redis-elasticache",
            "required": [
                "url",
                "username",
                "password",
            ],
        },
        "redshift": {
            "name": "redshift",
            "required": [
                "connection_url",
            ],
        },
        "snowflake": {
            "name": "snowflake",
            "required": [
                "connection_url",
            ],
        },
        "default": {
            "name": "",
            "required": [],
        },
    }
)


def get_plugin_meta(name):
    """
    Get meta information for a plugin with this name,
    excluding the `-database-plugin` suffix.
    """
    return PLUGINS.get(name, PLUGINS["default"])


def get_plugin_name(name):
    """
    Get the name of a plugin as rendered by this module. This is a utility for the state
    module primarily.
    """
    plugin_name = PLUGINS.get(name, {"name": name})["name"]
    return f"{plugin_name}-database-plugin"


def create_cache_pattern(name=None, mount=None, cache=None, static=None):
    """
    Render a match pattern for operating on cached leases.
    Unset parameters will result in a ``*`` glob.

    name
        The name of the database role.

    static
        Whether the role is static.

    cache
        Filter by cache name (refer to get_creds for details).

    mount
        The mount path the associated database backend is mounted to.
    """
    ptrn = ["db"]
    ptrn.append("*" if mount is None else mount)
    ptrn.append("*" if static is None else "static" if static else "dynamic")
    ptrn.append("*" if name is None else name)
    if cache is True or static is True:
        # Conceptually, there can only be one static lease since it's
        # tied to a specific account.
        ptrn.append("default")
    elif cache:
        ptrn.append(cache)
    else:
        ptrn.append("*")
    return ".".join(ptrn)



import copy
import time
from unittest.mock import ANY
from unittest.mock import Mock
from unittest.mock import patch

import pytest
import salt.cache

from saltext.vault.utils import vault
from saltext.vault.utils.vault import cache as vcache


@pytest.fixture
def cbank():
    return "vault/connection"


@pytest.fixture
def ckey():
    return "test"


@pytest.fixture
def data():
    return {"foo": "bar"}


@pytest.fixture
def context(cbank, ckey, data):
    return {cbank: {ckey: data}}


@pytest.fixture
def cached(cache_factory, data):
    cache = Mock(spec=salt.cache.Cache)
    cache.contains.return_value = True
    cache.fetch.return_value = data
    cache.updated.return_value = time.time()
    cache_factory.return_value = cache
    return cache


@pytest.fixture
def cached_outdated(cache_factory, data):
    cache = Mock(spec=salt.cache.Cache)
    cache.contains.return_value = True
    cache.fetch.return_value = data
    cache.updated.return_value = time.time() - 9999999
    cache_factory.return_value = cache
    return cache


@pytest.fixture
def uncached(cache_factory):
    cache = Mock(spec=salt.cache.Cache)
    cache.contains.return_value = False
    cache.fetch.return_value = None
    cache.updated.return_value = None
    cache_factory.return_value = cache
    return cache


@pytest.fixture(autouse=True, params=[0])
def time_stopped(request):
    with patch(
        "saltext.vault.utils.vault.cache.time.time",
        autospec=True,
        return_value=request.param,
    ):
        yield


@pytest.mark.usefixtures("salt_runtype")
@pytest.mark.parametrize("connection", [True, False])
@pytest.mark.parametrize(
    "salt_runtype,force_local,expected",
    [
        ("MASTER", False, "vault"),
        ("MASTER_IMPERSONATING", False, "minions/test-minion/vault"),
        ("MASTER_IMPERSONATING", True, "vault"),
        ("MINION_LOCAL", False, "vault"),
        ("MINION_REMOTE", False, "vault"),
    ],
    indirect=["salt_runtype"],
)
def test_get_cache_bank(connection, force_local, expected):
    """
    Ensure the cache banks are mapped as expected, depending on run type
    """
    opts = {"grains": {"id": "test-minion"}}
    cbank = vcache._get_cache_bank(opts, force_local=force_local, connection=connection)
    if connection:
        expected += "/connection"
    assert cbank == expected


class TestVaultCache:
    """
    Vault Cache tests
    """

    @pytest.mark.parametrize("config", ["session", "other"])
    def test_get_uncached(self, config, uncached, cbank, ckey):
        """
        Ensure that unavailable cached data is reported as None.
        """
        cache = vcache.VaultCache(
            {}, cbank, ckey, cache_backend=uncached if config != "session" else None
        )
        res = cache.get()
        assert res is None
        if config != "session":
            uncached.contains.assert_called_once_with(cbank, ckey)

    def test_get_cached_from_context(self, context, cached, cbank, ckey, data):
        """
        Ensure that cached data in __context__ is respected, regardless
        of cache backend.
        """
        cache = vcache.VaultCache(context, cbank, ckey, cache_backend=cached)
        res = cache.get()
        assert res == data
        cached.updated.assert_not_called()
        cached.fetch.assert_not_called()

    def test_get_cached_not_outdated(self, cached, cbank, ckey, data):
        """
        Ensure that cached data that is still valid is returned.
        """
        cache = vcache.VaultCache({}, cbank, ckey, cache_backend=cached, ttl=3600)
        res = cache.get()
        assert res == data
        cached.updated.assert_called_once_with(cbank, ckey)
        cached.fetch.assert_called_once_with(cbank, ckey)

    def test_get_cached_outdated(self, cached_outdated, cbank, ckey):
        """
        Ensure that cached data that is not valid anymore is flushed
        and None is returned by default.
        """
        cache = vcache.VaultCache({}, cbank, ckey, cache_backend=cached_outdated, ttl=1)
        res = cache.get()
        assert res is None
        cached_outdated.updated.assert_called_once_with(cbank, ckey)
        cached_outdated.flush.assert_called_once_with(cbank, ckey)
        cached_outdated.fetch.assert_not_called()

    @pytest.mark.parametrize("config", ["session", "other"])
    def test_flush(self, config, context, cached, cbank, ckey):
        """
        Ensure that flushing clears the context key only and, if
        a cache backend is in use, it is also cleared.
        """
        cache = vcache.VaultCache(
            context, cbank, ckey, cache_backend=cached if config != "session" else None
        )
        cache.flush()
        assert context == {cbank: {}}
        if config != "session":
            cached.flush.assert_called_once_with(cbank, ckey)

    @pytest.mark.parametrize("config", ["session", "other"])
    def test_flush_cbank(self, config, context, cached, cbank, ckey):
        """
        Ensure that flushing with cbank=True clears the context bank and, if
        a cache backend is in use, it is also cleared.
        """
        cache = vcache.VaultCache(
            context, cbank, ckey, cache_backend=cached if config != "session" else None
        )
        cache.flush(cbank=True)
        assert context == {}
        if config != "session":
            cached.flush.assert_called_once_with(cbank, None)

    @pytest.mark.parametrize("context", [{}, {"vault/connection": {}}])
    @pytest.mark.parametrize("config", ["session", "other"])
    def test_store(self, config, context, uncached, cbank, ckey, data):
        """
        Ensure that storing data in cache always updates the context
        and, if a cache backend is in use, it is also stored there.
        """
        cache = vcache.VaultCache(
            context,
            cbank,
            ckey,
            cache_backend=uncached if config != "session" else None,
        )
        cache.store(data)
        assert context == {cbank: {ckey: data}}
        if config != "session":
            uncached.store.assert_called_once_with(cbank, ckey, data)


class TestVaultConfigCache:
    """
    Tests for Vault Config Cache
    """

    @pytest.fixture(params=["session", "other", None])
    def config(self, request):
        if request.param is None:
            return None
        return {
            "cache": {
                "backend": request.param,
                "config": 3600,
                "secret": "ttl",
            }
        }

    @pytest.fixture
    def data(self):
        return {
            "cache": {
                "backend": "new",
                "config": 1337,
                "secret": "ttl",
            }
        }

    @pytest.mark.usefixtures("uncached")
    def test_get_config_cache_uncached(self, cbank, ckey):
        """
        Ensure an uninitialized instance is returned when there is no cache
        """
        res = vault.cache._get_config_cache({}, {}, cbank, ckey)
        assert res.config is None

    def test_get_config_context_cached(self, uncached, cbank, ckey, context):
        """
        Ensure cached data in context wins
        """
        res = vault.cache._get_config_cache({}, context, cbank, ckey)
        assert res.config == context[cbank][ckey]
        uncached.contains.assert_not_called()

    def test_get_config_other_cached(self, cached, cbank, ckey, data):
        """
        Ensure cached data from other sources is respected
        """
        res = vault.cache._get_config_cache({}, {}, cbank, ckey)
        assert res.config == data
        cached.contains.assert_called_once_with(cbank, ckey)
        cached.fetch.assert_called_once_with(cbank, ckey)

    def test_reload(self, config, data, cbank, ckey):
        """
        Ensure that a changed configuration is reloaded correctly and
        during instantiation. When the config backend changes and the
        previous was not session only, it should be flushed.
        """
        with patch("saltext.vault.utils.vault.cache.VaultConfigCache.flush") as flush:
            cache = vcache.VaultConfigCache({}, cbank, ckey, {}, init_config=config)
            assert cache.config == config
            if config is not None:
                assert cache.ttl == config["cache"]["config"]
                if config["cache"]["backend"] != "session":
                    assert cache.cache is not None
            else:
                assert cache.ttl is None
                assert cache.cache is None
            cache._load(data)
            assert cache.ttl == data["cache"]["config"]
            assert cache.cache is not None
            if config is not None and config["cache"]["backend"] != "session":
                flush.assert_called_once()

    @pytest.mark.usefixtures("cached")
    def test_exists(self, config, context, cbank, ckey):
        """
        Ensure exists always evaluates to false when uninitialized
        """
        cache = vcache.VaultConfigCache(context, cbank, ckey, {}, init_config=config)
        res = cache.exists()
        assert res is bool(config)

    def test_get(self, config, cached, context, cbank, ckey, data):
        """
        Ensure cached data is returned and backend settings honored,
        unless the instance has not been initialized yet
        """
        if config is not None and config["cache"]["backend"] != "session":
            context = {}
        cache = vcache.VaultConfigCache(context, cbank, ckey, {}, init_config=config)
        res = cache.get()
        if config is not None:
            assert res == data
            if config["cache"]["backend"] != "session":
                cached.fetch.assert_called_once_with(cbank, ckey)
            else:
                cached.contains.assert_not_called()
                cached.fetch.assert_not_called()
        else:
            # uninitialized should always return None
            # initialization when first stored or constructed with init_config
            cached.contains.assert_not_called()
            assert res is None

    def test_flush(self, config, context, cached, cbank, ckey):
        """
        Ensure flushing deletes the whole cache bank (=connection scope),
        unless the configuration has not been initialized.
        Also, it should uninitialize the instance.
        """
        if config is None:
            context_old = copy.deepcopy(context)
        cache = vcache.VaultConfigCache(context, cbank, ckey, {}, init_config=config)
        cache.flush()
        if config is None:
            assert context == context_old
            cached.flush.assert_not_called()
        else:
            if config["cache"]["backend"] == "session":
                assert context == {}
            else:
                cached.flush.assert_called_once_with(cbank, None)
            assert cache.ttl is None
            assert cache.cache is None
            assert cache.config is None

    @pytest.mark.usefixtures("uncached")
    def test_store(self, data, cbank, ckey):
        """
        Ensure storing config in cache also reloads the instance
        """
        cache = vcache.VaultConfigCache({}, {}, cbank, ckey)
        assert cache.config is None
        with patch("saltext.vault.utils.vault.cache.VaultConfigCache._load") as rld:
            with patch("saltext.vault.utils.vault.cache.VaultCache.store") as store:
                cache.store(data)
                rld.assert_called_once_with(data)
                store.assert_called_once()

    @pytest.mark.parametrize("config", ["other"], indirect=True)
    def test_flush_exceptions_with_flush(self, config, cached, cbank, ckey):
        """
        Ensure internal flushing is disabled when the object is initialized
        with a reference to an exception class.
        """
        cache = vcache.VaultConfigCache(
            {},
            cbank,
            ckey,
            {},
            cache_backend_factory=lambda *args: cached,
            flush_exception=vault.VaultConfigExpired,
            init_config=config,
        )
        with pytest.raises(vault.VaultConfigExpired):
            cache.flush()

    @pytest.mark.parametrize("config", ["other"], indirect=True)
    def test_flush_exceptions_with_get(self, config, cached_outdated, cbank, ckey):
        """
        Ensure internal flushing is disabled when the object is initialized
        with a reference to an exception class.
        """
        cache = vcache.VaultConfigCache(
            {},
            cbank,
            ckey,
            {},
            cache_backend_factory=lambda *args: cached_outdated,
            flush_exception=vault.VaultConfigExpired,
            init_config=config,
        )
        with pytest.raises(vault.VaultConfigExpired):
            cache.get()


class TestVaultAuthCache:
    """
    Tests for Vault Auth Cache
    """

    @pytest.fixture
    def uncached(self):
        with patch(
            "saltext.vault.utils.vault.cache.CommonCache._ckey_exists",
            return_value=False,
            autospec=True,
        ):
            with patch(
                "saltext.vault.utils.vault.cache.CommonCache._get_ckey",
                return_value=None,
                autospec=True,
            ) as get:
                yield get

    @pytest.fixture
    def cached(self, token_auth):
        with patch(
            "saltext.vault.utils.vault.cache.CommonCache._ckey_exists",
            return_value=True,
            autospec=True,
        ):
            with patch(
                "saltext.vault.utils.vault.cache.CommonCache._get_ckey",
                return_value=token_auth["auth"],
                autospec=True,
            ) as get:
                yield get

    @pytest.fixture
    def cached_outdated(self, token_auth):
        with patch(
            "saltext.vault.utils.vault.cache.CommonCache._ckey_exists",
            return_value=True,
            autospec=True,
        ):
            token_auth["auth"]["creation_time"] = 0
            token_auth["auth"]["lease_duration"] = 1
            with patch(
                "saltext.vault.utils.vault.cache.CommonCache._get_ckey",
                return_value=token_auth["auth"],
                autospec=True,
            ) as get:
                yield get

    @pytest.fixture
    def cached_invalid_flush(self, token_auth, cached):
        with patch("saltext.vault.utils.vault.cache.CommonCache._flush", autospec=True) as flush:
            token_auth["auth"]["num_uses"] = 1
            token_auth["auth"]["use_count"] = 1
            cached.return_value = token_auth["auth"]
            yield flush

    @pytest.mark.usefixtures("uncached")
    def test_get_uncached(self):
        """
        Ensure that unavailable cached data is reported as None.
        """
        cache = vcache.VaultAuthCache({}, "cbank", "ckey", vault.VaultToken)
        res = cache.get()
        assert res is None

    @pytest.mark.usefixtures("cached")
    def test_get_cached(self, token_auth):
        """
        Ensure that cached data that is still valid is returned.
        """
        cache = vcache.VaultAuthCache({}, "cbank", "ckey", vault.VaultToken)
        res = cache.get()
        assert res is not None
        assert res == vault.VaultToken(**token_auth["auth"])

    def test_get_cached_invalid(self, cached_invalid_flush):
        """
        Ensure that cached data that is not valid anymore is flushed
        and None is returned.
        """
        cache = vcache.VaultAuthCache({}, "cbank", "ckey", vault.VaultToken)
        res = cache.get()
        assert res is None
        cached_invalid_flush.assert_called_once()

    def test_store(self, token_auth):
        """
        Ensure that storing authentication data sends a dictionary
        representation to the store implementation of the parent class.
        """
        token = vault.VaultToken(**token_auth["auth"])
        cache = vcache.VaultAuthCache({}, "cbank", "ckey", vault.VaultToken)
        with patch("saltext.vault.utils.vault.cache.CommonCache._store_ckey") as store:
            cache.store(token)
            store.assert_called_once_with("ckey", token.to_dict())

    def test_flush_exceptions_with_flush(self, cached, cbank, ckey):
        """
        Ensure internal flushing is disabled when the object is initialized
        with a reference to an exception class.
        """
        cache = vcache.VaultAuthCache(
            {},
            cbank,
            ckey,
            vault.VaultToken,
            cache_backend=cached,
            flush_exception=vault.VaultAuthExpired,
        )
        with pytest.raises(vault.VaultAuthExpired):
            cache.flush()

    @pytest.mark.usefixtures("cached_outdated")
    def test_flush_exceptions_with_get(self, cbank, ckey):
        """
        Ensure internal flushing is disabled when the object is initialized
        with a reference to an exception class.
        """
        cache = vcache.VaultAuthCache(
            {}, cbank, ckey, vault.VaultToken, flush_exception=vault.VaultAuthExpired
        )
        with pytest.raises(vault.VaultAuthExpired):
            cache.get(10)


class TestVaultLeaseCache:
    """
    Tests for Vault Lease Cache
    """

    @pytest.fixture
    def uncached(self):
        with patch(
            "saltext.vault.utils.vault.cache.CommonCache._ckey_exists",
            return_value=False,
            autospec=True,
        ):
            with patch(
                "saltext.vault.utils.vault.cache.CommonCache._get_ckey",
                return_value=None,
                autospec=True,
            ) as get:
                yield get

    @pytest.fixture
    def cached(self, lease):
        with patch(
            "saltext.vault.utils.vault.cache.CommonCache._ckey_exists",
            return_value=True,
            autospec=True,
        ):
            with patch(
                "saltext.vault.utils.vault.cache.CommonCache._get_ckey",
                return_value=lease,
                autospec=True,
            ) as get:
                yield get

    @pytest.fixture
    def cached_outdated(self, lease):
        with patch(
            "saltext.vault.utils.vault.cache.CommonCache._ckey_exists",
            return_value=True,
            autospec=True,
        ):
            lease["duration"] = 6
            lease["expire_time"] = 6
            with patch(
                "saltext.vault.utils.vault.cache.CommonCache._get_ckey",
                return_value=lease,
                autospec=True,
            ) as get:
                yield get

    @pytest.mark.usefixtures("uncached")
    def test_get_uncached(self):
        """
        Ensure that unavailable cached data is reported as None.
        """
        cache = vcache.VaultLeaseCache({}, "cbank")
        res = cache.get("testlease")
        assert res is None

    @pytest.mark.usefixtures("cached")
    def test_get_cached(self, lease):
        """
        Ensure that cached data that is still valid is returned.
        """
        cache = vcache.VaultLeaseCache({}, "cbank")
        res = cache.get("testlease")
        assert res is not None
        assert res == vault.VaultLease(**lease)

    @pytest.mark.usefixtures("cached", "time_stopped")
    @pytest.mark.parametrize("valid_for,expected", ((1, True), (99999999, False)))
    def test_get_cached_valid_for(self, valid_for, expected, lease):
        """
        Ensure that requesting leases with a validity works as expected.
        The lease should be returned if it is valid, otherwise only
        the invalid ckey should be flushed and None returned.
        """
        cache = vcache.VaultLeaseCache({}, "cbank")
        with patch(
            "saltext.vault.utils.vault.cache.CommonCache._flush",
            autospec=True,
        ) as flush:
            res = cache.get("testlease", valid_for=valid_for, flush=True)
            if expected:
                flush.assert_not_called()
                assert res is not None
                assert res == vault.VaultLease(**lease)
            else:
                flush.assert_called_once_with(ANY, "testlease")
                assert res is None

    def test_store(self, lease):
        """
        Ensure that storing authentication data sends a dictionary
        representation to the store implementation of the parent class.
        """
        lease_ = vault.VaultLease(**lease)
        cache = vcache.VaultLeaseCache({}, "cbank")
        with patch("saltext.vault.utils.vault.cache.CommonCache._store_ckey") as store:
            cache.store("ckey", lease_)
            store.assert_called_once_with("ckey", lease_.to_dict())

    @pytest.mark.parametrize("lease", ({}, {"meta": {"foo": "bar"}}), indirect=True)
    @pytest.mark.usefixtures("cached_outdated")
    def test_expire_events_with_get(self, events, lease):
        """
        Ensure internal flushing is disabled when the object is initialized
        with a reference to an exception class.
        """
        cache = vcache.VaultLeaseCache({}, "cbank", expire_events=events)
        ret = cache.get("ckey", 10)
        assert ret is None
        events.assert_called_once_with(
            tag="vault/lease/ckey/expire",
            data={"valid_for_less": 10, "ttl_left": 6, "meta": lease["meta"]},
        )



"""
Manage the Vault (or OpenBao) database secret engine, request and cache
leased database credentials.

.. versionadded:: 1.1.0

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import logging

from salt.defaults import NOT_SET
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

from saltext.vault.utils.vault import db as vaultdb
from saltext.vault.utils.vault.helpers import timestring_map

log = logging.getLogger(__name__)


def connection_present(
    name,
    plugin,
    version=None,
    verify=True,
    allowed_roles=None,
    root_rotation_statements=None,
    password_policy=None,
    rotate=True,
    force=False,
    mount="database",
    **kwargs,
):
    """
    Ensure a database connection is present as specified.

    name
        The name of the database connection.

    plugin
        The name of the database plugin. Known plugins to this module are:
        ``cassandra``, ``couchbase``, ``elasticsearch``, ``influxdb``, ``hanadb``, ``mongodb``,
        ``mongodb_atlas``, ``mssql``, ``mysql``, ``oracle``, ``postgresql``, ``redis``,
        ``redis_elasticache``, ``redshift``, ``snowflake``.
        If you pass an unknown plugin, make sure its Vault-internal name can be formatted
        as ``{plugin}-database-plugin`` and to pass all required parameters as kwargs.

    version
        Specifies the semantic version of the plugin to use for this connection.

    verify
        Verify the connection during initial configuration. Defaults to True.

    allowed_roles
        List of the roles allowed to use this connection. ``["*"]`` means any role
        can use this connection. Defaults to empty (no role can use it).

    root_rotation_statements
        Specifies the database statements to be executed to rotate the root user's credentials.
        See the plugin's API page for more information on support and formatting for this parameter.

    password_policy
        The name of the password policy to use when generating passwords for this database.
        If not specified, this will use a default policy defined as:
        20 characters with at least 1 uppercase, 1 lowercase, 1 number, and 1 dash character.

    rotate
        Rotate the root credentials after plugin setup. Defaults to True.

    force
        When the plugin changes, this state fails to protect from accidental errors.
        Set force to True to delete existing connections with the same name and a
        different plugin type. Defaults to False.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.

    kwargs
        Different plugins require different parameters. You need to make sure that you pass them
        as supplemental keyword arguments. For known plugins, the required arguments will
        be checked.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": "The connection is present as specified",
        "changes": {},
    }
    kwargs = {k: v for k, v in kwargs.items() if not k.startswith("_")}

    def _diff_params(current):
        nonlocal version, allowed_roles, root_rotation_statements, password_policy, kwargs
        diff_params = (
            ("plugin_version", version),
            ("allowed_roles", allowed_roles),
            ("root_credentials_rotate_statements", root_rotation_statements),
            ("password_policy", password_policy),
            # verify_connection is not reported
        )
        changed = {}
        for param, arg in diff_params:
            if arg is None:
                continue
            # Strip statements to avoid tripping over final newlines
            if param.endswith("statements"):
                arg = [x.rstrip() for x in arg]
                if param in current:
                    current[param] = [x.rstrip() for x in current[param]]
            if param not in current or current[param] != arg:
                changed.update({param: {"old": current.get(param), "new": arg}})
        for param, val in kwargs.items():
            if param == "password":
                # password is not reported
                continue
            if (
                param not in current["connection_details"]
                or current["connection_details"][param] != val
            ):
                changed.update(
                    {param: {"old": current["connection_details"].get(param), "new": val}}
                )
        return changed

    try:
        current = __salt__["vault_db.fetch_connection"](name, mount=mount)
        changes = {}

        if current:
            if current["plugin_name"] != vaultdb.get_plugin_name(plugin):
                if not force:
                    raise CommandExecutionError(
                        "Cannot change plugin type without deleting the existing connection. "
                        "Set force: true to override."
                    )
                if not __opts__["test"]:
                    __salt__["vault_db.delete_connection"](name, mount=mount)
                ret["changes"]["deleted_for_plugin_change"] = name
                current = None
            else:
                changes = _diff_params(current)
                if not changes:
                    return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = (
                f"Connection `{name}` would have been {'updated' if current else 'created'}"
            )
            ret["changes"].update(changes)
            if not current:
                ret["changes"]["created"] = name
            return ret

        if current and "password" in kwargs:
            kwargs.pop("password")

        __salt__["vault_db.write_connection"](
            name,
            plugin,
            version=version,
            verify=verify,
            allowed_roles=allowed_roles,
            root_rotation_statements=root_rotation_statements,
            password_policy=password_policy,
            rotate=rotate,
            mount=mount,
            **kwargs,
        )
        new = __salt__["vault_db.fetch_connection"](name, mount=mount)

        if new is None:
            raise CommandExecutionError(
                "There were no errors during role management, but it is still reported as absent."
            )
        if not current:
            ret["changes"]["created"] = name

        new_diff = _diff_params(new)
        if new_diff:
            ret["result"] = False
            ret["comment"] = (
                "There were no errors during connection management, but "
                f"the reported parameters do not match: {new_diff}"
            )
            return ret
        ret["changes"].update(changes)
        ret["comment"] = f"Connection `{name}` has been {'updated' if current else 'created'}"

    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = str(err)
        # do not reset changes

    return ret


def connection_absent(name, mount="database"):
    """
    Ensure a database connection is absent.

    name
        The name of the connection.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    try:
        current = __salt__["vault_db.fetch_connection"](name, mount=mount)

        if current is None:
            ret["comment"] = f"Connection `{name}` is already absent."
            return ret

        ret["changes"]["deleted"] = name

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Connection `{name}` would have been deleted."
            return ret

        __salt__["vault_db.delete_connection"](name, mount=mount)

        if __salt__["vault_db.fetch_connection"](name, mount=mount) is not None:
            raise CommandExecutionError(
                "There were no errors during connection deletion, "
                "but it is still reported as present."
            )
        ret["comment"] = f"Connection `{name}` has been deleted."

    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def role_present(
    name,
    connection,
    creation_statements,
    default_ttl=None,
    max_ttl=None,
    revocation_statements=None,
    rollback_statements=None,
    renew_statements=None,
    credential_type=None,
    credential_config=None,
    mount="database",
):
    """
    Ensure a regular database role is present as specified.

    name
        The name of the database role.

    connection
        The name of the database connection this role applies to.

    creation_statements
        Specifies a list of database statements executed to create and configure a user,
        usually templated with {{name}} and {{password}}. Required.

    default_ttl
        Specifies the TTL for the leases associated with this role. Accepts time suffixed
        strings (1h) or an integer number of seconds. Defaults to system/engine default TTL time.

    max_ttl
        Specifies the maximum TTL for the leases associated with this role. Accepts time suffixed
        strings (1h) or an integer number of seconds. Defaults to sys/mounts's default TTL time;
        this value is allowed to be less than the mount max TTL (or, if not set,
        the system max TTL), but it is not allowed to be longer.

    revocation_statements
        Specifies a list of database statements to be executed to revoke a user.

    rollback_statements
        Specifies a list of database statements to be executed to rollback a create operation
        in the event of an error. Availability and formatting depend on the specific plugin.

    renew_statements
        Specifies a list of database statements to be executed to renew a user.
        Availability and formatting depend on the specific plugin.

    credential_type
        Specifies the type of credential that will be generated for the role.
        Options include: ``password``, ``rsa_private_key``. Defaults to ``password``.
        See the plugin's API page for credential types supported by individual databases.

    credential_config
        Specifies the configuration for the given ``credential_type`` as a mapping.
        For ``password``, only ``password_policy`` can be passed.
        For ``rsa_private_key``, ``key_bits`` (defaults to 2048) and ``format``
        (defaults to ``pkcs8``) are available.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if not isinstance(creation_statements, list):
        creation_statements = [creation_statements]
    if revocation_statements and not isinstance(revocation_statements, list):
        revocation_statements = [revocation_statements]
    if rollback_statements and not isinstance(rollback_statements, list):
        rollback_statements = [rollback_statements]
    if renew_statements and not isinstance(renew_statements, list):
        renew_statements = [renew_statements]

    def _diff_params(current):
        nonlocal connection, creation_statements, default_ttl, max_ttl, revocation_statements
        nonlocal rollback_statements, renew_statements, credential_type, credential_config

        diff_params = (
            ("db_name", connection),
            ("creation_statements", creation_statements),
            ("default_ttl", timestring_map(default_ttl)),
            ("max_ttl", timestring_map(max_ttl)),
            ("revocation_statements", revocation_statements),
            ("rollback_statements", rollback_statements),
            ("renew_statements", renew_statements),
            ("credential_type", credential_type),
            ("credential_config", credential_config),
        )
        changed = {}
        for param, arg in diff_params:
            if arg is None:
                continue
            # Strip statements to avoid tripping over final newlines
            if param.endswith("statements"):
                arg = [x.rstrip() for x in arg]
                if param in current:
                    current[param] = [x.rstrip() for x in current[param]]
            if param not in current or current[param] != arg:
                changed.update({param: {"old": current.get(param), "new": arg}})
        return changed

    try:
        current = __salt__["vault_db.fetch_role"](name, static=False, mount=mount)

        if current:
            changed = _diff_params(current)
            if not changed:
                ret["comment"] = "Role is present as specified"
                return ret
            ret["changes"].update(changed)

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Role `{name}` would have been {'updated' if current else 'created'}"
            if not current:
                ret["changes"]["created"] = name
            return ret

        __salt__["vault_db.write_role"](
            name,
            connection,
            creation_statements,
            default_ttl=default_ttl,
            max_ttl=max_ttl,
            revocation_statements=revocation_statements,
            rollback_statements=rollback_statements,
            renew_statements=renew_statements,
            credential_type=credential_type,
            credential_config=credential_config,
            mount=mount,
        )
        new = __salt__["vault_db.fetch_role"](name, static=False, mount=mount)

        if new is None:
            raise CommandExecutionError(
                "There were no errors during role management, but it is still reported as absent."
            )

        if not current:
            ret["changes"]["created"] = name

        new_diff = _diff_params(new)
        if new_diff:
            ret["result"] = False
            ret["comment"] = (
                "There were no errors during role management, but "
                f"the reported parameters do not match: {new_diff}"
            )
            return ret

        ret["comment"] = f"Role `{name}` has been {'updated' if current else 'created'}"
    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def role_absent(name, static=False, mount="database"):
    """
    Ensure a database role is absent.

    name
        The name of the role.

    static
        Whether this role is static. Defaults to False.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    try:
        current = __salt__["vault_db.fetch_role"](name, static=static, mount=mount)

        if current is None:
            ret["comment"] = f"Role `{name}` is already absent."
            return ret

        ret["changes"]["deleted"] = name

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Role `{name}` would have been deleted."
            return ret

        __salt__["vault_db.delete_role"](name, static=static, mount=mount)

        if __salt__["vault_db.fetch_role"](name, static=static, mount=mount) is not None:
            raise CommandExecutionError(
                "There were no errors during role deletion, but it is still reported as present."
            )
        ret["comment"] = f"Role `{name}` has been deleted."

    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def static_role_present(
    name,
    connection,
    username,
    rotation_period,
    rotation_statements=None,
    credential_type=None,
    credential_config=None,
    mount="database",
):
    """
    Ensure a database Static Role is present as specified.

    name
        The name of the database role.

    connection
        The name of the database connection this role applies to.

    username
        The username to manage.

    rotation_period
        Specifies the amount of time Vault should wait before rotating the password.
        The minimum is ``5s``.

    rotation_statements
        Specifies the database statements to be executed to rotate the password for the
        configured database user. Not every plugin type will support this functionality.

    credential_type
        Specifies the type of credential that will be generated for the role.
        Options include: ``password``, ``rsa_private_key``. Defaults to ``password``.
        See the plugin's API page for credential types supported by individual databases.

    credential_config
        Specifies the configuration for the given ``credential_type`` as a mapping.
        For ``password``, only ``password_policy`` can be passed.
        For ``rsa_private_key``, ``key_bits`` (defaults to 2048) and ``format``
        (defaults to ``pkcs8``) are available.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if rotation_statements and not isinstance(rotation_statements, list):
        rotation_statements = [rotation_statements]

    def _diff_params(current):
        nonlocal connection, username, rotation_period, rotation_statements, credential_type, credential_config
        diff_params = (
            ("db_name", connection),
            ("username", username),
            ("rotation_period", timestring_map(rotation_period)),
            ("rotation_statements", rotation_statements),
            ("credential_type", credential_type),
            ("credential_config", credential_config),
        )
        changed = {}
        for param, arg in diff_params:
            if arg is None:
                continue
            # Strip statements to avoid tripping over final newlines
            if param.endswith("statements"):
                arg = [x.rstrip() for x in arg]
                if param in current:
                    current[param] = [x.rstrip() for x in current[param]]
            if param not in current or current[param] != arg:
                changed.update({param: {"old": current.get(param), "new": arg}})
        return changed

    try:
        current = __salt__["vault_db.fetch_role"](name, static=True, mount=mount)

        if current:
            changed = _diff_params(current)
            if not changed:
                ret["comment"] = "Role is present as specified"
                return ret
            ret["changes"].update(changed)

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Role `{name}` would have been {'updated' if current else 'created'}"
            if not current:
                ret["changes"]["created"] = name
            return ret

        __salt__["vault_db.write_static_role"](
            name,
            connection,
            username,
            rotation_period,
            rotation_statements=None,
            credential_type=credential_type,
            credential_config=credential_config,
            mount=mount,
        )
        new = __salt__["vault_db.fetch_role"](name, static=True, mount=mount)

        if new is None:
            raise CommandExecutionError(
                "There were no errors during role management, but it is still reported as absent."
            )

        if not current:
            ret["changes"]["created"] = name

        new_diff = _diff_params(new)
        if new_diff:
            ret["result"] = False
            ret["comment"] = (
                "There were no errors during role management, but "
                f"the reported parameters do not match: {new_diff}"
            )
            return ret

        ret["comment"] = f"Role `{name}` has been {'updated' if current else 'created'}"

    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def creds_cached(
    name,
    static=False,
    cache=None,
    valid_for=NOT_SET,
    renew_increment=None,
    revoke_delay=None,
    meta=None,
    mount="database",
    **kwargs,  # pylint: disable=unused-argument
):
    """
    Ensure valid credentials are present in the minion's cache based on the named role.
    Supports ``mod_beacon``.

    .. note::

        This function is mostly intended to associate a specific credential with
        a beacon that warns about expiry and allows to run an associated state to
        reconfigure an application with new credentials.
        See the :py:mod:`vault_lease beacon module <saltext.vault.beacons.vault_lease>`
        for an :ref:`example state to configure a lease together with a beacon <beacon-state-example>`.

    name
        The name of the database role.

    static
        Whether this role is static. Defaults to False.

    cache
        A variable cache suffix to be able to use multiple distinct credentials
        using the same role on the same minion.
        Ignored when ``static`` is true.

        .. note::

            This uses the same cache backend as the Vault integration, so make
            sure you configure a persistent backend like ``disk`` if you expect
            the credentials to survive a single run.

    valid_for
        Ensure the credentials are valid for at least this amount of time,
        otherwise request new ones.
        This can be an integer, which will be interpreted as seconds, or a time string
        using the same format as Vault does:
        Suffix ``s`` for seconds, ``m`` for minuts, ``h`` for hours, ``d`` for days.
        Defaults to ``0``.

    renew_increment
        When using cache and ``valid_for`` results in a renewal attempt, request this
        amount of time extension on the lease. This will be cached together with the
        lease and might be used by other modules later.

    revoke_delay
        When using cache and ``valid_for`` results in a revocation, set the lease
        validity to this value to allow a short amount of delay between the issuance
        of the new lease and the revocation of the old one. Defaults to ``60``.
        This will be cached together with the lease and might be used by other
        modules later.

    meta
        When using cache, this value will be cached together with the lease. It will
        be emitted by the ``vault_lease`` beacon module whenever a lease is
        running out (usually because it cannot be extended further). It is intended
        to support the reactor in deciding what needs to be done in order
        to to reconfigure dependent, Vault-unaware software with newly issued
        credentials. Entirely optional.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": "The credentials are already cached and valid",
        "changes": {},
    }

    cached = __salt__["vault_db.list_cached"](
        name, static=static, cache=cache or True if not static else True, mount=mount
    )
    pp = "issued"
    if cached:
        info = cached[next(iter(cached))]
        if valid_for is NOT_SET:
            if info["min_ttl"] is not None:
                valid_for = info["min_ttl"]
            else:
                valid_for = None
        for attr, val in (
            ("min_ttl", valid_for),
            ("renew_increment", renew_increment),
            ("revoke_delay", revoke_delay),
            ("meta", meta),
        ):
            if val is not None and info.get(attr) != val:
                # Meta-only changes should be reported as well because the
                # execution module needs to be called later to update them.
                # This is especially valid for a lowering of min_ttl, which
                # might result in a reissuance if the current lease has already
                # reached its min_ttl (the current logic would not recognize that
                # situation otherwise).
                ret["changes"][attr] = {"old": info.get(attr), "new": val}
                pp = "edited"

        current_effective_valid_for = valid_for or 0
        if info["min_ttl"] is not None:
            current_effective_valid_for = max(info["min_ttl"], valid_for or 0)
        if info["expires_in"] <= timestring_map(current_effective_valid_for):
            ret["changes"]["expiry"] = True
            pp = "renewed"
        if not ret["changes"]:
            return ret
    else:
        ret["changes"]["new"] = True
    if __opts__["test"]:
        ret["result"] = None
        if pp == "renewed":
            pp = "renewed/reissued"
        ret["comment"] = f"The credentials would have been {pp}"
        return ret
    __salt__["vault_db.get_creds"](
        name,
        static=static,
        cache=cache or True,
        valid_for=valid_for,
        renew_increment=renew_increment,
        revoke_delay=revoke_delay,
        meta=meta,
        mount=mount,
        _warn_about_attr_change=False,
    )
    new_cached = __salt__["vault_db.list_cached"](name, static=static, cache=cache, mount=mount)
    if not new_cached:
        raise CommandExecutionError(
            "Could not find cached credentials after issuing, this is likely a bug"
        )
    # Ensure the reporting is correct.
    if cached and new_cached[next(iter(cached))]["lease_id"] != info["lease_id"]:
        pp = "reissued"
        ret["changes"][pp] = True

    ret["comment"] = f"The credentials have been {pp}"
    return ret


def creds_uncached(
    name, static=False, cache=None, mount="database", **kwargs
):  # pylint: disable=unused-argument
    """
    Ensure credentials are absent in the minion's cache based on the named role.
    Supports ``mod_beacon``.

    .. note::

        This function is mostly intended to remove a cached lease and its
        beacon. See :py:func:`creds_cached` for a more detailed description.
        To remove the associated beacon together with the lease, just pass
        ``beacon: true`` as a parameter to this state.

    name
        The name of the database role.

    static
        Whether this role is static. Defaults to False.

    cache
        A variable cache suffix to be able to use multiple distinct credentials
        using the same role on the same minion.
        Ignored when ``static`` is true.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": "No matching credentials present",
        "changes": {},
    }

    cached = __salt__["vault_db.list_cached"](name, static=static, cache=cache, mount=mount)
    if not cached:
        return ret
    ret["changes"]["revoked"] = True
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "The credentials would have been revoked"
        return ret
    __salt__["vault_db.clear_cached"](name, static=static, cache=cache or True, mount=mount)
    ret["comment"] = "The credentials have been revoked"
    return ret


def mod_beacon(name, sfun=None, static=False, cache=None, mount="database", **kwargs):
    """
    Associates a Vault lease with a ``vault_lease`` beacon and
    possibly a state.

    beacon_interval
        The interval to run the beacon in. Defaults to 60.

    min_ttl
        If this minimum TTL on the lease is undercut, the beacon will
        fire an event. Defaults to 0.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    supported_funcs = ("creds_cached", "creds_uncached")

    if sfun not in supported_funcs:
        ret["result"] = False
        ret["comment"] = f"'vault_db.{sfun}' does not work with mod_beacon"
        return ret
    if not kwargs.get("beacon"):
        ret["comment"] = "Not managing beacon"
        return ret
    lease = vaultdb.create_cache_pattern(name, mount=mount, static=static, cache=cache or True)
    beacon_module = "vault_lease"
    beacon_name = f"{beacon_module}_{lease}"
    if sfun == "creds_uncached":
        beacon_kwargs = {
            "name": beacon_name,
            "beacon_module": beacon_module,
        }
        bfun = "absent"
    else:
        # sfun == "creds_cached" evidently
        beacon_kwargs = {
            "name": beacon_name,
            "beacon_module": beacon_module,
            "interval": kwargs.get("beacon_interval", 60),
            "leases": lease,
            "min_ttl": kwargs.get("min_ttl", 0),
            "meta": kwargs.get("meta"),
            "check_server": kwargs.get("check_server", False),
        }
        bfun = "present"
    return __states__[f"beacon.{bfun}"](**beacon_kwargs)



import re

from . import prompt
from .cmd import ProcessExecutionError
from .cmd import git
from .cmd import local
from .git import list_untracked

PRE_COMMIT_TEST_REGEX = re.compile(
    r"^(?P<test>[^\n]+?)\.{4,}.*(?P<resolution>Failed|Passed|Skipped)$"
)
NON_IDEMPOTENT_HOOKS = (
    "trim trailing whitespace",
    "mixed line ending",
    "fix end of files",
    "Remove Python Import Header Comments",
    "Check rST doc files exist for modules/states",
    "Salt extensions docstrings auto-fixes",
    "Rewrite the test suite",
    "Rewrite Code to be Py3.",
    "isort",
    "black",
    "blacken-docs",
)


def parse_pre_commit(data):
    """
    Parse pre-commit output into a list of passing hooks and a mapping of
    failing hooks to their output.
    """
    passing = []
    failing = {}
    cur = None
    for line in data.splitlines():
        if match := PRE_COMMIT_TEST_REGEX.match(line):
            cur = None
            if match.group("resolution") != "Failed":
                passing.append(match.group("test"))
                continue
            cur = match.group("test")
            failing[cur] = []
            continue
        try:
            failing[cur].append(line)
        except KeyError:
            # in case the parsing logic fails, let's not crash everything
            continue
    return passing, {test: "\n".join(output).strip() for test, output in failing.items()}


def check_pre_commit_rerun(data):
    """
    Check if we can expect failing hooks to turn green during a rerun.
    """
    _, failing = parse_pre_commit(data)
    for hook in failing:
        if hook.startswith(NON_IDEMPOTENT_HOOKS):
            return True
    return False


def run_pre_commit(venv, retries=2):
    """
    Run pre-commit in a loop until it passes, there is no chance of
    autoformatting to make it pass or a maximum number of runs is reached.

    Usually, a maximum of two runs is necessary (if a hook reformats the
    output of another later one again).
    """
    new_files = set()

    def _run_pre_commit_loop(retries_left):
        untracked_files = set(map(str, list_untracked()))
        nonlocal new_files
        new_files = new_files.union(untracked_files)
        # Ensure pre-commit runs on all paths.
        # We don't want to git add . because this removes merge conflicts
        git("add", "--intent-to-add", *untracked_files)
        with local.venv(venv):
            try:
                local["python"]("-m", "pre_commit", "run", "--all-files")
            except ProcessExecutionError as err:
                if retries_left > 0 and check_pre_commit_rerun(err.stdout):
                    return _run_pre_commit_loop(retries_left - 1)
                raise

    prompt.status(
        "Running pre-commit hooks against all files. This can take a minute, please be patient"
    )

    try:
        _run_pre_commit_loop(retries)
        return True
    except ProcessExecutionError as err:
        _, failing = parse_pre_commit(err.stdout)
        if failing:
            msg = f"Please fix all ({len(failing)}) failing hooks"
        else:
            msg = f"Output: {err.stderr or err.stdout}"
        prompt.warn(f"Pre-commit is failing. {msg}")
        for i, failing_hook in enumerate(failing):
            prompt.warn(f"✗ Failing hook ({i + 1}): {failing_hook}", failing[failing_hook])
    finally:
        if new_files:
            try:
                # Check if there is at least one commit in the repo,
                # otherwise git restore --staged fails.
                git("rev-parse", "HEAD")
            except ProcessExecutionError:
                pass
            else:
                # Undo git add --intent-to-add to allow RenovateBot to detect new files correctly
                git("restore", "--staged", *new_files)
    return False



import pytest
import salt.version
from packaging.version import Version


@pytest.fixture(scope="module", autouse=True)
def _check_host_python(salt_ssh_cli):
    """
    When testing Salt 3006.*, the host's default Python version
    needs to be <3.12, otherwise Salt just crashes with an ImportError
    regarding ``backports.ssl_match_hostname``.
    """
    if Version(salt.version.__version__) >= Version("3007"):
        return
    ret = salt_ssh_cli.run("--raw", "python3 --version")
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    python_version = Version(ret.data["stdout"].split(" ")[1])
    if python_version >= Version("3.12"):
        pytest.skip(
            f"The host Python ({python_version}) is not supported by Salt {salt.version.__version__}"
        )



"""
Helper for warning about deprecations
"""

import inspect
import os
import sys
import warnings

import packaging.version

from saltext.vault import __version__


def warn_until(
    version,
    message,
    category=DeprecationWarning,
):
    """
    Warn about deprecations until the specified version is reached, after which
    raise a RuntimeError to remind developers about removal.

    Loosely based on ``salt.utils.versions.warn_until``.

    version
        The version at which the warning turns into an error. Can be specified
        as a string, float, int or iterable with items castable to integers.

    message
        The warning message to show.

    category
        The warning class to be thrown, by default ``DeprecationWarning``.
    """
    version = _parse_version(version)
    saltext_version = _parse_version(__version__)
    # Attribute the warning to the calling function, not to warn_until()
    stacklevel = 2

    if saltext_version >= version:
        caller = inspect.getframeinfo(sys._getframe(stacklevel - 1))
        raise RuntimeError(
            f"The warning triggered on filename '{caller.filename}', line number "
            f"{caller.lineno}, is supposed to be shown until version "
            f"{version} is released. Current version is now "
            f"{saltext_version}. Please remove the warning."
        )

    if os.environ.get("PYTHONWARNINGS") != "ignore":
        warnings.warn(
            message.format(version=version),
            category,
            stacklevel=stacklevel,
        )


class Version(packaging.version.Version):
    def __lt__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__lt__(other)

    def __le__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__le__(other)

    def __eq__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__eq__(other)

    def __ge__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__ge__(other)

    def __gt__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__gt__(other)

    def __ne__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__ne__(other)


def _parse_version(version):
    if isinstance(version, str):
        pass
    elif isinstance(version, (float, int)):
        version = str(version)
    else:
        try:
            version = ".".join(str(x) for x in version)
        except TypeError as err:
            raise RuntimeError("`version` must be a string, integer, float or an iterable") from err
    return Version(version)



import contextlib
from unittest.mock import patch

import pytest
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

from saltext.vault.modules import vault_db
from saltext.vault.utils import vault


@pytest.fixture
def configure_loader_modules():
    return {
        vault_db: {
            "__grains__": {"id": "test-minion"},
        }
    }


@pytest.fixture
def _conn_absent():
    with patch(
        "saltext.vault.modules.vault_db.fetch_connection", return_value=None, autospec=True
    ) as fetch:
        yield fetch


@pytest.fixture
def query():
    with patch("saltext.vault.utils.vault.query", return_value=True, autospec=True) as _query:
        yield _query


@pytest.mark.parametrize(
    "func,kwargs",
    (
        ("list_connections", {}),
        ("fetch_connection", {"name": "foo"}),
        ("write_connection", {"name": "foo", "plugin": "custom"}),
        ("delete_connection", {"name": "foo"}),
        ("reset_connection", {"name": "foo"}),
        ("rotate_root", {"name": "foo"}),
        ("list_roles", {}),
        ("fetch_role", {"name": "foo"}),
        (
            "write_static_role",
            {"name": "foo", "connection": "bar", "username": "baz", "rotation_period": 42},
        ),
        (
            "write_role",
            {"name": "foo", "connection": "bar", "creation_statements": "thou shall exist"},
        ),
        ("delete_role", {"name": "foo"}),
        ("get_creds", {"name": "foo", "cache": False}),
        ("rotate_static_role", {"name": "foo"}),
    ),
)
def test_func_converts_errors(func, kwargs, query, request):
    query.side_effect = vault.VaultException("booh")
    if func == "write_connection":
        # otherwise we would test fetch_connection again
        request.getfixturevalue("_conn_absent")
    with pytest.raises(CommandExecutionError, match="booh"):
        getattr(vault_db, func)(**kwargs)


@pytest.mark.usefixtures("_conn_absent")
@pytest.mark.parametrize("plugin", ("mysql", "custom"))
def test_write_connection_missing_kwargs(plugin):
    if plugin == "custom":
        ctx = patch("saltext.vault.utils.vault.query", autospec=True)
    else:
        ctx = pytest.raises(SaltInvocationError, match="requires.*additional.*connection_url")
    with ctx:
        vault_db.write_connection("foo", plugin)


@pytest.mark.usefixtures("_conn_absent")
def test_write_connection_payload(query):
    kwargs = {
        "version": "1.2.3",
        "verify": True,
        "allowed_roles": ["*"],
        "root_rotation_statements": ["rotate!"],
        "password_policy": "yolo",
        "custom_arg": True,
    }
    assert vault_db.write_connection("foo", "custom", **kwargs, rotate=False, mount="bar") is True
    endpoint = query.call_args[0][1]
    payload = query.call_args[1]["payload"]
    assert endpoint == "bar/config/foo"
    expected_payload = kwargs.copy()
    expected_payload["plugin_name"] = "custom-database-plugin"
    expected_payload["plugin_version"] = expected_payload.pop("version")
    expected_payload["verify_connection"] = expected_payload.pop("verify")
    assert payload == expected_payload


@pytest.mark.usefixtures("_conn_absent")
@pytest.mark.parametrize("rotate", (False, True))
def test_write_connection_rotate(query, rotate):
    vault_db.write_connection("foo", "custom", rotate=rotate)
    endpoint = query.call_args[0][1]
    assert (endpoint == "database/config/foo") is not rotate
    assert (endpoint == "database/rotate-root/foo") is rotate


def test_write_static_role_payload(query):
    kwargs = {
        "rotation_period": 42,
        "rotation_statements": ["rotate!"],
        "credential_type": "password",
        "credential_config": {"password_policy": "yolo"},
    }
    assert vault_db.write_static_role("role", "conn", "user", **kwargs, mount="mount") is True
    endpoint = query.call_args[0][1]
    payload = query.call_args[1]["payload"]
    assert endpoint == "mount/static-roles/role"
    expected_payload = kwargs.copy()
    expected_payload["username"] = "user"
    expected_payload["db_name"] = "conn"
    assert payload == expected_payload


def test_write_role_payload(query):
    kwargs = {
        "creation_statements": ["cogito ergo sum"],
        "default_ttl": 42,
        "max_ttl": 1337,
        "revocation_statements": ["it's not you, it's me"],
        "rollback_statements": ["this should be fine"],
        "renew_statements": ["kekkon shitemo kudasai"],
        "credential_type": "rsa_private_key",
        "credential_config": {"key_bits": 1},
    }
    assert vault_db.write_role("role", "conn", **kwargs, mount="mount") is True
    endpoint = query.call_args[0][1]
    payload = query.call_args[1]["payload"]
    assert endpoint == "mount/roles/role"
    expected_payload = kwargs.copy()
    expected_payload["db_name"] = "conn"
    assert payload == expected_payload


@pytest.mark.parametrize(
    "typ,vals,expected",
    (
        (None, None, True),
        (None, {"password_policy": "yolo"}, True),
        (None, {"password_police": "??"}, False),
        (None, {"key_bits": 1}, False),
        ("password", {"password_policy": "yolo"}, True),
        ("password", {"password_alice": "257"}, False),
        ("password", {"key_bits": 1}, False),
        ("rsa_private_key", {"key_bits": 1, "format": "red"}, True),
        ("rsa_private_key", {"key_fits": 0}, False),
        ("rsa_private_key", {"password_policy": "yolo"}, False),
        ("unknown", {"something": "else"}, True),
    ),
)
@pytest.mark.usefixtures("query")
def test_write_role_credential_type_param_verification(typ, vals, expected):
    if expected:
        ctx = contextlib.nullcontext()
    else:
        ctx = pytest.raises(SaltInvocationError, match="invalid for credential type")
    with ctx:
        vault_db.write_static_role(
            "role", "conn", "user", 42, credential_type=typ, credential_config=vals
        )



import pytest

from tests.support.vault import vault_delete_policy
from tests.support.vault import vault_delete_secret
from tests.support.vault import vault_list_policies
from tests.support.vault import vault_read_secret
from tests.support.vault import vault_write_policy
from tests.support.vault import vault_write_secret

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def master_config_overrides():
    """
    You can override the default configuration per package by overriding this
    fixture in a conftest.py file.
    """
    return {
        "vault": {
            "cache": {
                "backend": "disk",
            },
            "policies": {
                "assign": [
                    "salt_minion",
                    "policy_admin",
                ],
            },
        }
    }


@pytest.fixture
def existing_secret():
    key = "secret/foo"
    vault_write_secret(key, user="foo", password="bar")
    try:
        yield key
    finally:
        # ensure we don't retain the version history between tests
        vault_delete_secret(key, metadata=True)


@pytest.fixture
def existing_secret_version(existing_secret):
    vault_write_secret(existing_secret, user="foo", password="hunter1")
    yield existing_secret


def test_read_secret(salt_ssh_cli, existing_secret):
    ret = salt_ssh_cli.run("vault.read_secret", existing_secret)
    assert ret.returncode == 0
    assert ret.data == {"user": "foo", "password": "bar"}


def test_read_secret_version(salt_ssh_cli, existing_secret_version):
    ret = salt_ssh_cli.run("vault.read_secret", existing_secret_version, version=1)
    assert ret.returncode == 0
    assert ret.data == {"user": "foo", "password": "bar"}
    ret = salt_ssh_cli.run("vault.read_secret", existing_secret_version, version=2)
    assert ret.returncode == 0
    assert ret.data == {"user": "foo", "password": "hunter1"}


def test_write_secret(salt_ssh_cli, existing_secret):
    ret = salt_ssh_cli.run("vault.write_secret", existing_secret, bar="baz")
    assert ret.returncode == 0
    new = vault_read_secret(existing_secret)
    assert new == {"bar": "baz"}


def test_write_raw(salt_ssh_cli, existing_secret):
    ret = salt_ssh_cli.run("vault.write_raw", existing_secret, {"bar": "baz"})
    assert ret.returncode == 0
    new = vault_read_secret(existing_secret)
    assert new == {"bar": "baz"}


def test_patch_secret(salt_ssh_cli, existing_secret):
    ret = salt_ssh_cli.run("vault.patch_secret", existing_secret, bar="baz")
    assert ret.returncode == 0
    new = vault_read_secret(existing_secret)
    assert new == {"user": "foo", "password": "bar", "bar": "baz"}


def test_delete_restore_secret(salt_ssh_cli, existing_secret):
    ret = salt_ssh_cli.run("vault.delete_secret", existing_secret)
    assert ret.returncode == 0
    assert vault_read_secret(existing_secret) is None
    ret = salt_ssh_cli.run("vault.restore_secret", existing_secret)
    assert ret.returncode == 0
    assert vault_read_secret(existing_secret)


@pytest.mark.usefixtures("existing_secret_version")
def test_delete_restore_secret_all_versions(salt_ssh_cli, existing_secret):
    versions = ("bar", "hunter1")
    ret = salt_ssh_cli.run("vault.delete_secret", existing_secret, all_versions=True)
    assert ret.returncode == 0
    assert ret.data is True
    for version in range(len(versions)):
        ret = salt_ssh_cli.run(
            "vault.read_secret", existing_secret, default="__was_deleted__", version=version + 1
        )
        assert ret.returncode == 0
        assert ret.data == "__was_deleted__"
    ret = salt_ssh_cli.run("vault.restore_secret", existing_secret, all_versions=True)
    assert ret.returncode == 0
    assert ret.data is True
    for version, data in enumerate(versions):
        ret = salt_ssh_cli.run("vault.read_secret", existing_secret, version=version + 1)
        assert ret.returncode == 0
        assert ret.data["password"] == data


def test_destroy_secret(salt_ssh_cli, existing_secret_version):
    ret = salt_ssh_cli.run("vault.destroy_secret", existing_secret_version)
    assert ret.returncode == 0
    assert vault_read_secret(existing_secret_version) is None
    ret = salt_ssh_cli.run("vault.read_secret", existing_secret_version, version=1)
    assert ret.returncode == 0
    assert ret.data["password"] == "bar"


def test_wipe_secret(salt_ssh_cli, existing_secret_version):
    ret = salt_ssh_cli.run("vault.wipe_secret", existing_secret_version)
    assert ret.returncode == 0
    assert vault_read_secret(existing_secret_version) is None
    ret = salt_ssh_cli.run("vault.read_secret_meta", existing_secret_version)
    assert ret.returncode == 0
    assert ret.data is False


@pytest.mark.usefixtures("existing_secret")
def test_list_secrets(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault.list_secrets", "secret", keys_only=True)
    assert ret.returncode == 0
    assert ret.data == ["foo"]


def test_clear_cache(salt_ssh_cli):
    """
    Ensure that the revocation client also respects Salt-SSH master opts
    and does not crash with InvalidConfig.
    """
    ret = salt_ssh_cli.run("vault.query", "GET", "auth/token/lookup-self")
    assert ret.returncode == 0
    token_id = ret.data["data"]["id"]
    ret = salt_ssh_cli.run("vault.query", "GET", "auth/token/lookup-self")
    assert ret.returncode == 0
    assert ret.data["data"]["id"] == token_id
    ret = salt_ssh_cli.run("vault.clear_cache")
    assert ret.returncode == 0
    assert ret.data is True
    ret = salt_ssh_cli.run("vault.query", "GET", "auth/token/lookup-self")
    assert ret.returncode == 0
    assert ret.data["data"]["id"] != token_id


def test_policies_list(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault.policies_list")
    assert ret.returncode == 0
    assert isinstance(ret.data, list)
    for policy in ("database_admin", "policy_admin", "salt_master", "salt_minion"):
        assert policy in ret.data


@pytest.fixture
def _cleanup_dummy_policy():
    yield
    vault_delete_policy("foobar")


@pytest.fixture
def _dummy_policy_contents():
    return """
path "foobar/*" {
    capabilities = ["read", "create", "update", "delete", "list", "patch"]
}
    """.strip()


@pytest.fixture
def _dummy_policy(_dummy_policy_contents, _cleanup_dummy_policy):  # pylint: disable=unused-argument
    name = "foobar"
    vault_write_policy(name, _dummy_policy_contents)
    yield name


@pytest.mark.usefixtures("_cleanup_dummy_policy")
def test_policy_write(salt_ssh_cli, _dummy_policy_contents):
    assert "foobar" not in vault_list_policies()
    ret = salt_ssh_cli.run("vault.policy_write", "foobar", _dummy_policy_contents)
    assert ret.returncode == 0
    assert ret.data is True
    assert "foobar" in vault_list_policies()


def test_policy_fetch(salt_ssh_cli, _dummy_policy, _dummy_policy_contents):
    ret = salt_ssh_cli.run("vault.policy_fetch", _dummy_policy)
    assert ret.returncode == 0
    assert ret.data == _dummy_policy_contents


def test_policy_delete(salt_ssh_cli, _dummy_policy):
    ret = salt_ssh_cli.run("vault.policy_delete", _dummy_policy)
    assert ret.returncode == 0
    assert ret.data is True
